###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Utilities for scripts dealing with transformations."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
import datetime
import random
import time
import sys
import six

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.List import breakListIntoChunks
# from DIRAC.Core.Utilities.Time import timeThis
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers, resolveSEGroup
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient

from DIRAC.TransformationSystem.Client.Utilities import PluginUtilities as DIRACPluginUtilities
from DIRAC.TransformationSystem.Client.Utilities import isArchive, getActiveSEs

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient, BKClientWithRetry
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery, makeBKPath
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.DataManagementSystem.Client.DMScript import ProgressBar
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

__RCSID__ = "$Id$"


def _stripNumDirectory(dirName):
  """Method to remove numeric directory form a directory path."""
  while True:
    subDir = os.path.basename(dirName)
    if subDir and not subDir.isdigit():
      return dirName
    dirName = os.path.dirname(dirName)


def _clearTaskLFNs(taskLfns):
  """Clear the input list of lists, keeping its reference."""
  nbLfns = 0
  for lfnList in taskLfns:
    nbLfns += len(lfnList)
    del lfnList[:]
  return nbLfns


class PluginUtilities(DIRACPluginUtilities):
  """Utility class used by plugins."""

  def __init__(self, plugin='LHCbStandard', transClient=None, dataManager=None, fc=None,
               bkClient=None, rmClient=None,
               debug=False, transInThread=None, transID=None):
    """c'tor."""
    # clients
    if transClient is None:
      self.transClient = TransformationClient()
    else:
      self.transClient = transClient
    if bkClient is None:
      bkClient = BookkeepingClient()
    self.bkClient = BKClientWithRetry(bkClient)
    if rmClient is None:
      self.rmClient = ResourceManagementClient()
    else:
      self.rmClient = rmClient

    super(PluginUtilities, self).__init__(plugin=plugin, transClient=self.transClient,
                                          dataManager=dataManager, fc=fc,
                                          debug=debug, transInThread=transInThread, transID=transID)

    self.freeSpace = {}
    self.transFiles = []
    self.transReplicas = {}
    self.cachedLFNAncestors = {}
    self.cachedNbRAWFiles = {}
    self.usingRAWFiles = None
    self.cachedRunLfns = {}
    self.cachedProductions = {}
    self.cachedLastRun = 0
    self.cachedLFNProcessedPath = {}
    self.cacheFile = ''
    self.filesParam = {}
    self.transRunFiles = {}
    self.notProcessed = {}
    self.cacheHitFrequency = max(0., 1 - self.getPluginParam('RunCacheUpdateFrequency', 0.1))
    self.__runExpired = {}
    self.__recoType = ''
    self.dmsHelper = DMSHelpers()
    self.runDestinations = {}
    self.cachedDirMetadata = {}
    self.runsUsedForShares = set()
    self.shareMetrics = None
    self.lastCall = None
    self.excludedStatuses = self.getPluginParam('IgnoreStatusForFlush', ['Removed', 'MissingInFC', 'Problematic'])
    self.tsParams = ('FileSize', 'FileType', 'RunNumber')
    self.paramName = None
    self.onlineCondDB = None
    self.chunkForFileDescendants = None
    # Method aliases
    self.setCachedTimeExceeded = self.setCachedLastRun
    self.getCachedTimeExceeded = self.getCachedLastRun

  def setDebug(self, val):
    self.debug = val

  def updateSharesUsage(self, counters, se, count, runID):
    """Update the usage counters if share is by files, and the run duration
    otherwise."""
    inc = 0
    if self.shareMetrics == 'Files':
      inc = count
    else:
      self.logVerbose("Updating SE %s for run %d which is %salready used" %
                      (se, runID, 'not ' if runID not in self.runsUsedForShares else ''))
      if runID not in self.runsUsedForShares:
        # Update with run duration
        res = self.__getRunDuration(runID)
        if not res['OK']:
          self.logError("Error getting run time", res['Message'])
        else:
          inc = res['Value']
          self.runsUsedForShares.add(runID)
    if inc:
      counters[se] = counters.setdefault(se, 0) + inc
      self.logVerbose('New %s counter for %s: %d (+%d)' % (self.shareMetrics, se, counters[se], inc))
      self.printShares("New counters and used fraction (%)", counters, log=self.logVerbose)

  def printShares(self, title, shares, counters=None, log=None):
    """Print formatted shares and if provised counters.

    :param title: header of print out
    :type title: string
    :param shares: Shares per SE or site
    :type shares: dictionary
    :param counters: Current values per SE or site
    :type counters: dictionary
    :param log: (sub)Logger instance to be used
    :type log: logging method
    """
    if log is None:
      log = self.logInfo
    if counters is None:
      counters = shares
    normCounters = normaliseShares(counters) if counters else {}

    if self.shareMetrics:
      title += " (metrics is %s)" % self.shareMetrics
    log(title)
    for se in sorted(shares):
      if isinstance(shares[se], six.integer_types):
        infoStr = "%s: %d " % (se.rjust(15), shares[se])
      else:
        infoStr = "%s: %4.1f " % (se.rjust(15), shares[se])
      if se in counters:
        infoStr += "| %4.1f %%" % normCounters[se]
      log(infoStr)

  def getPluginShares(self, section=None, backupSE=None, log=None, counters=True):
    """Get shares from CS:

    * If backupSE is not present: just return the CS shares as they are
    * If backupSE is specified, the shares represent a percentage of the RAW at each site and the rest is for backupSE

    If the ShareMetrics is Files:
    * Use the number of files as metrics for SE usage
    Else:
    * Use the duration (in seconds) of runs as metrics for SE usage
    """
    if log is None:
      log = self.logInfo
    if section is None:
      # This is for reconstruction shares (CU)
      sharesSections = {'DataReconstruction': 'CPUforRAW', 'DataReprocessing': 'CPUforReprocessing'}
      res = self.transClient.getTransformation(self.transID)
      if not res['OK']:
        self.logError("Cannot get information on transformation")
        return res
      else:
        transType = res['Value']['Type']
      section = sharesSections.get(transType, 'CPUforRAW')
    res = getShares(section)
    if not res['OK']:
      self.logError("There is no CS section %s" % section, res['Message'])
      return res
    rawFraction = {}
    if backupSE:
      # Apply these processing fractions to the RAW distribution shares
      rawFraction = res['Value']
      result = getShares('RAW', normalise=True)
      if result['OK']:
        rawShares = result['Value']
        shares = dict((se, rawShares[se] * rawFraction[se]) for se in set(rawShares) & set(rawFraction))
        tier1Fraction = sum(shares.values())
        shares[backupSE] = 100. - tier1Fraction
      else:
        return res
      rawPercentage = dict((se, 100. * val) for se, val in rawFraction.items())  # can be an iterator
      self.printShares("Fraction of RAW (%s) to be processed at each SE (%%):" % section,
                       rawPercentage, counters=[], log=log)
    else:
      shares = normaliseShares(res['Value'])
      log("Obtained the following target distribution shares (%):")
      for se in sorted(shares):
        log("%s: %.1f" % (se.ljust(15), shares[se]))

    # If counters are not requested, return only shares
    if not counters:
      return S_OK(({}, shares))

    if self.shareMetrics is None:
      self.shareMetrics = self.getPluginParam('ShareMetrics', '')
    # Get the existing destinations from the transformationDB, just for printing
    if self.shareMetrics == 'Files':
      res = self.getExistingCounters(requestedSEs=sorted(shares))
    else:
      res = self.getSitesRunsDuration(requestedSEs=sorted(shares))
    if not res['OK']:
      self.logError("Failed to get used share", res['Message'])
      return res
    else:
      existingCount = res['Value']
      self.printShares("Target shares and usage for production (%):", shares, counters=existingCount, log=log)
    if rawFraction:
      return S_OK((rawFraction, shares))
    else:
      return S_OK((existingCount, shares))

  def __getRunDuration(self, runID):
    """Get the run duration."""
    # Get run metadata
    runMetadata = self.transClient.getRunsMetadata([runID]).get('Value', {})
    return S_OK(self.__extractRunDuration(runMetadata, runID))

  def __extractRunDuration(self, runMetadata, runID):
    """Obtain run duration from the TS run metadata table If not available, get
    it from BK."""
    duration = runMetadata.get(runID, {}).get('Duration')
    if duration is None:
      self.logVerbose('Run duration not found in TS for run %d, get it from BK' % runID)
      res = self.bkClient.getRunInformation({'RunNumber': [runID], 'Fields': ['JobStart', 'JobEnd']})
      if not res['OK']:
        self.logError("Error getting run start/end information", res['Message'])
        duration = 0
      if runID in res['Value']:
        start = res['Value'][runID]['JobStart']
        end = res['Value'][runID]['JobEnd']
        duration = (end - start).seconds
      else:
        duration = 0
    return int(duration)

  def getSitesRunsDuration(self, transID=None, normalise=False, requestedSEs=None):
    """Get per site how much time of run was assigned."""
    res = self.getTransformationRuns(transID=transID)
    if not res['OK']:
      return res
    runDictList = res['Value']
    # Get run metadata
    runMetadata = self.transClient.getRunsMetadata([runDict['RunNumber'] for runDict in runDictList]).get('Value', {})

    seUsage = {}
    for runDict in runDictList:
      runID = runDict['RunNumber']
      selectedSEs = runDict['SelectedSite']
      selectedSEs = set(selectedSEs.split(',')) if selectedSEs is not None else set()
      if requestedSEs:
        selectedSEs &= set(requestedSEs)
        if not selectedSEs:
          continue
      duration = self.__extractRunDuration(runMetadata, runID)
      self.runsUsedForShares.add(runID)
      for se in selectedSEs:
        seUsage[se] = seUsage.setdefault(se, 0) + duration

    if normalise:
      seUsage = normaliseShares(seUsage)
    return S_OK(seUsage)

  def getExistingCounters(self, transID=None, normalise=False, requestedSEs=None):
    """Used by RAWReplication and RAWProcessing plugins, gets what has been
    done up to now while distributing runs."""
    if transID is None:
      transID = self.transID
    res = self.transClient.getCounters('TransformationFiles', ['UsedSE'],
                                       {'TransformationID': transID})
    if not res['OK']:
      return res
    usageDict = {}
    for usedDict, count in res['Value']:
      usedSE = usedDict['UsedSE']
      if usedSE != 'Unknown':
        usageDict[usedSE] = count
    if requestedSEs:
      requestedSEs = set(requestedSEs)
      seDict = {}
      for se, count in usageDict.items():  # can be an iterator
        overlap = set(se.split(',')) & requestedSEs
        if overlap:
          for ov in overlap:
            seDict[ov] = seDict.setdefault(ov, 0) + count
        else:
          self.logWarn("%s is in counters but not in required list" % se)
      usageDict = seDict
    if normalise:
      usageDict = normaliseShares(usageDict)
    return S_OK(usageDict)

  def getMetadataFromTSorBK(self, lfns, param):
    """Get BK parameters from BK unless they are already present in the TS."""
    if isinstance(lfns, six.string_types):
      lfns = [lfns]
    filesParam = {}
    # For parameters that are in the TS files DB, no need to go to BK
    if param in self.tsParams:
      par = param if param != 'FileSize' else 'Size'
      # We already have this information
      for fileDict in self.transFiles:
        lfn = fileDict['LFN']
        if lfn in lfns:
          value = fileDict.get(par)
          if value:
            filesParam[lfn] = value

      lfnsLeft = set(lfns) - set(filesParam)
      self.logVerbose("Found parameter %s in transFiles for %d files, %d left" % (param,
                                                                                  len(filesParam),
                                                                                  len(lfnsLeft)))
    else:
      lfnsLeft = lfns

    if lfnsLeft:
      for chunk in breakListIntoChunks(lfnsLeft, 1000):
        loggedError = False
        # Let's loop forever as this MUST succeed
        while True:
          res = self.bkClient.getFileMetadata(chunk)
          if res['OK']:
            success = res['Value']['Successful']
            filesParam.update(dict((lfn, success[lfn].get(param)) for lfn in success))
            break
          if not loggedError:
            self.logError("Error getting BK file metadata", res['Message'])
            loggedError = True
          time.sleep(1)
      # Check if all parameters are present in TS... If not, set them while we have them here
      lfnDict = {}
      for fileDict in self.transFiles:
        lfn = fileDict['LFN']
        if lfn in success:
          for param in self.tsParams:
            par = param if param != 'FileSize' else 'Size'
            if fileDict.get(par) is None:
              lfnDict.setdefault(lfn, {}).update({par: success[lfn][param]})
      if lfnDict:
        res = self.transClient.setParameterToTransformationFiles(self.transID, lfnDict)
        self.logVerbose("Updated files metadata in TS for %d files: %s" % (len(lfnDict),
                                                                           'OK' if res['OK'] else 'Failed'))

    return filesParam

  def getProductions(self, processingPasses, transStatus):
    """Get the list of productions matching a given list of processing
    passes."""
    # Get information about the current transformation
    transQuery = self.getTransQuery(self.transReplicas)
    if transQuery is None:
      return S_ERROR("Could not get transformation BK query")
    transProcPass = transQuery['ProcessingPass']
    eventType = transQuery['EventType']
    if not transProcPass:
      self.logError('Unable to find processing pass for transformation')
      return S_ERROR('No processing pass found')

    # Build list of BK paths to look for
    fullBKPaths = [bkPath for bkPath in processingPasses if bkPath.startswith('/')]
    relBKPaths = [bkPath for bkPath in processingPasses if not bkPath.startswith('/')]
    bkPathList = {}
    if fullBKPaths:
      # We were given full BK paths, use them
      bkQuery = BKQuery(bkPath, visible='All')
      eventType = bkQuery.getEventTypeList()
      bkPathList.update(dict((bkPath.replace('RealData', 'Real Data'), bkQuery.getQueryDict()['ProcessingPass'])
                             for bkPath in fullBKPaths))

    if relBKPaths:
      for procPass in relBKPaths:
        if not transProcPass.endswith(procPass):
          newPass = os.path.join(transProcPass, procPass)
        else:
          newPass = procPass
        transQuery.update({'ProcessingPass': newPass, 'Visibility': 'All'})
        transQuery.pop('FileType', None)
        bkPathList[makeBKPath(transQuery)] = newPass
    self.logVerbose('List of BK paths:', '\n\t'.join([''] + ['%s: %s' %
                                                             (bkPath.replace('Real Data', 'RealData'),
                                                              bkPathList[bkPath].replace('Real Data', 'RealData'))
                                                             for bkPath in sorted(bkPathList)]))

    now = datetime.datetime.utcnow()
    period = self.getPluginParam('Period', 6)
    cacheLifeTime = self.getPluginParam('CacheLifeTime', 24)
    productions = self.getCachedProductions()
    # Check if the cache lifetime expired
    if 'CacheTime' in productions and (now - productions['CacheTime']) < datetime.timedelta(hours=cacheLifeTime):
      # If we haven't found productions for one of the processing passes, retry
      cacheOK = len([bkPath for bkPath in bkPathList if bkPath not in productions.get('List', {})]) == 0
    else:
      cacheOK = False

    if cacheOK:
      # Check if it's worth processing (i.e. "period" hours have passed
      if transStatus != 'Flush' and not self.getCachedTimeExceeded() and \
              (now - productions['LastCall_%s' % self.transID]) < datetime.timedelta(hours=period):
        self.logInfo("Skip this loop (less than %s hours since last call)" % period)
        return S_OK()
    else:
      productions.setdefault('List', {})
      self.logVerbose("Cache is being refreshed (lifetime %d hours)" % cacheLifeTime)
      for bkPath in bkPathList:
        bkQuery = BKQuery(bkPath, visible='All')
        bkQuery.setOption('EventType', eventType)
        bkQuery.setOption('ReplicaFlag', 'All')
        prods = bkQuery.getBKProductions()
        if not prods:
          self.logVerbose("For bkPath %s, found no productions, wait next time" % (bkPath))
          return S_OK()
        productions['List'][bkPath] = []
        for prod in prods:
          res = self.transClient.getTransformation(prod)
          if not res['OK']:
            self.logError("Error getting transformation %s" % prod, res['Message'])
          else:
            if res['Value']['Status'] != "Cleaned":
              productions['List'][bkPath].append(int(prod))
        self.logInfo("For bkPath %s, selected productions: %s" %
                     (bkPath, ','.join('%s' % prod for prod in productions['List'][bkPath])))
      productions['CacheTime'] = now

    # Record time of last call and update the cache
    productions['LastCall_%s' % self.transID] = now
    self.setCachedProductions(productions)
    # Return the list of BK paths (with their processing passes),
    #   the list of productions and the depth of current processing pass
    return S_OK((bkPathList, productions))

  # @timeThis
  def getFilesParam(self, lfns, param):
    """Return for each LFN the value of a single parameter "param" that can be
    cached."""
    # Make sure we request the same parameter
    if self.paramName and self.paramName != param:
      self.logWarn("Requested parameter %s not compatible with previously used parameter %s, \
get from BK" % (param, self.paramName))
      return self.getMetadataFromTSorBK(lfns, param)

    self.paramName = param
    lfns = set(lfns)
    if not self.filesParam:
      nCached = 0
      for run in self.cachedRunLfns:
        for paramValue in self.cachedRunLfns[run]:
          for lfn in self.cachedRunLfns[run][paramValue]:
            self.filesParam[lfn] = paramValue
            nCached += 1
      self.logVerbose('Obtained %s of %d files from disk cache' % (param, nCached))
    filesParam = dict((lfn, self.filesParam[lfn]) for lfn in lfns & set(self.filesParam))
    # Get parameter for LFNs not cached yet
    lfns = lfns - set(filesParam)
    if lfns:
      resDict = self.getMetadataFromTSorBK(list(lfns), param)
      filesParam.update(resDict)
      self.filesParam.update(resDict)
      self.logVerbose("Obtained %s of %d files from TS or BK" % (param, len(resDict)))
    return filesParam

  def getStorageFreeSpace(self, candSEs):
    """Get free space in a list of SEs from the RSS."""
    weight = {}
    for se in candSEs:
      weight[se] = self.getRMFreeSpace(se)
    self.logVerbose("Free space from RSS: %s" % weight)
    return weight

  def getRMFreeSpace(self, se):
    """Get free space in an SE from the RSS."""

    isTape = StorageElement(se).status()['TapeSE']
    # For tape space token, the SRM information is useless, set it fixed
    if isTape:
      return 1000.

    cacheLimit = datetime.datetime.utcnow() - datetime.timedelta(hours=12)

    if not (se in self.freeSpace) or self.freeSpace[se]['LastCheckTime'] < cacheLimit:
      res = self.rmClient.selectSpaceTokenOccupancyCache(token=se)
      if not res['OK'] or not res['Value']:
        self.logError('Error when getting space for SE %s' % (se,), res['Message'])
        return 0

      self.freeSpace[se] = dict(zip(res['Columns'], res['Value'][0]))

    # Return free space in TB as RSS returns in MB
    free = self.freeSpace[se]['Free'] / 1000000.
    self.logDebug('Free space for SE %s: %d' % (se, free))
    return free

  def rankSEs(self, candSEs):
    """Ranks the SEs according to their free space."""
    if len(candSEs) <= 1:
      return candSEs
    # Weights should be obtained from the RSS or CS
    weightForSEs = self.getStorageFreeSpace(candSEs)
    rankedSEs = []
    while weightForSEs:
      if len(weightForSEs) == 1:
        se = list(weightForSEs)[0]
      else:
        weights = weightForSEs.copy()
        total = 0.
        orderedSEs = []
        for se, weight in weights.items():  # can be an iterator
          # Minimum space 1 GB in case all are 0
          total += max(weight, 0.001)
          weights[se] = total
          orderedSEs.append(se)
        rand = random.uniform(0., total)
        self.logDebug('List of ordered SEs (random number is %.1f out of %.1f)' % (rand, total))
        for se in orderedSEs:
          self.logDebug('%s: %.1f' % (se, weights[se]))
        for se in orderedSEs:
          if rand <= weights[se]:
            break
      self.logDebug("Selected SE is %s" % se)
      rankedSEs.append(se)
      weightForSEs.pop(se)
    return rankedSEs

  def setTargetSEs(self, numberOfCopies, archiveSEs,
                   mandatorySEs, secondarySEs, existingSEs, exclusiveSEs=False):
    """Decide on which SEs to target from lists and current status of
    replication Policy is max one archive1, one archive 2, all mandatory SEs
    and required number of copies elsewhere."""
    # Select active SEs
    nbArchive = min(1, len(archiveSEs))
    secondaryActiveSEs = getActiveSEs(secondarySEs)

    targetSEs = []
    self.logVerbose("Selecting SEs from %s, %s, %s (%d copies) for files in %s" % (archiveSEs,
                                                                                   mandatorySEs,
                                                                                   secondarySEs,
                                                                                   numberOfCopies,
                                                                                   existingSEs))
    # Select archive SEs if any requested
    if nbArchive:
      allArchiveSEs = resolveSEGroup('Tier1-Archive')
      archiveExisting = [se for se in allArchiveSEs if se in existingSEs]
      archiveActiveSEs = getActiveSEs(archiveSEs)
      # If none found active, use all
      if not archiveActiveSEs:
        archiveActiveSEs = archiveSEs
      archiveActiveSEs = [se for se in archiveActiveSEs if se not in archiveExisting]
      # Set existing archive SEs first, then other archives (check against all archives)
      candidateSEs = self.selectSEs(archiveExisting + self.rankSEs(archiveActiveSEs), nbArchive, existingSEs)
      self.logVerbose("Selected ArchiveSEs: %s" % candidateSEs)
      if len(candidateSEs) < nbArchive:
        self.logError('Cannot select enough archive SEs')
        return None
      targetSEs += candidateSEs

    # Now select the disk replicas
    # 1. add mandatory SEs
    candidateSEs = [se for se in mandatorySEs if not self.isSameSEInList(se, existingSEs)]
    # 2. add existing disk SEs that are either mandatory or secondary
    candidateSEs += [
        se for se in existingSEs
        if se in (mandatorySEs + secondarySEs) and
        not self.isSameSEInList(se, targetSEs + candidateSEs) and
        not isArchive(se)
    ]
    # 3. add ranked list of secondary SEs
    candidateSEs += [
        se for se in self.rankSEs(secondaryActiveSEs)
        if not self.isSameSEInList(se, targetSEs + candidateSEs + existingSEs)
    ]
    # 4. Select the proper number of SEs in the candidate ordered list
    candidateSEs = self.selectSEs(candidateSEs, numberOfCopies, existingSEs)
    self.logVerbose("Selected SecondarySEs: %s" % candidateSEs)
    if len(candidateSEs) < numberOfCopies:
      self.logError("Can not select enough Active SEs as SecondarySE")
      return None
    targetSEs += candidateSEs

    # Remove existing SEs if requested
    if exclusiveSEs:
      targetSEs = [se for se in targetSEs if se not in existingSEs]
    self.logVerbose("Selected target SEs: %s" % targetSEs)
    return ','.join(sorted(targetSEs))

  def selectSEs(self, candSEs, needToCopy, existingSEs):
    """Select SEs from a list, preferably from existing SEs in order to obtain
    the required number of replicas."""
    targetSEs = []
    # Put existing SEs first
    for se in [se for se in candSEs if se in existingSEs]:
      if needToCopy <= 0:
        break
      targetSEs.append(se)
      needToCopy -= 1
    # If more are needed, add them
    if needToCopy > 0:
      for se in [s for s in candSEs if s not in existingSEs]:
        if needToCopy <= 0:
          break
        if not self.isSameSEInList(se, existingSEs):
          targetSEs.append(se)
          needToCopy -= 1
    return targetSEs

  def assignTargetToLfns(self, lfns, replicas, stringTargetSEs):
    """Assign target SEs for each LFN, excluding the existing ones Returns a
    dictionary for files to be transferred and a list of files already in
    place."""
    # Suppress duplicate SEs from list
    targetSEs = self.uniqueSEs([se for se in stringTargetSEs.split(',') if se])
    alreadyCompleted = []
    fileTargetSEs = {}
    for lfn in lfns:
      # At this stage, protection against Failover has been done
      existingSEs = replicas.get(lfn, [])
      if not existingSEs:
        self.logWarn('File found without replicas', lfn)
        continue
      # Discard existing SEs or SEs with similar description
      neededSEs = [se for se in targetSEs if not self.isSameSEInList(se, existingSEs)]
      if not neededSEs:
        alreadyCompleted.append(lfn)
      else:
        fileTargetSEs[lfn] = ','.join(sorted(neededSEs))
    return (fileTargetSEs, alreadyCompleted)

  # @timeThis
  def getProcessedFiles(self, lfns):
    """Check which files have been processed by a given production, i.e. have a
    meaningful descendant."""
    from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import getFileDescendants
    return getFileDescendants(self.transID, lfns, transClient=self.transClient,
                              dm=self.dm, bkClient=self.bkClient)

  # @timeThis
  def getRAWAncestorsForRun(self, runID, param=None, paramValue=None, getFiles=False):
    """Determine from BK how many ancestors files from a given run we have.

    This is used for deciding when to flush a run (when all RAW files
    have been processed)
    """
    ancestorFiles = set()
    # The transformation files cannot be cached globally as they evolve at each cycle
    lfns = self.transRunFiles.get(runID, [])
    if not lfns:
      res = self.getTransformationFiles(runID=runID)
      if not res['OK']:
        self.logError("Cannot get transformation files for run %s: %s" % (str(runID), res['Message']))
        return 0
      lfns = [fileDict['LFN'] for fileDict in res['Value'] if fileDict['Status'] not in self.excludedStatuses]
      self.transRunFiles[runID] = lfns
      self.logVerbose('Obtained %d input files for run %d' % (len(lfns), runID))

    if not lfns:
      return 0

    # Restrict to files with the required parameter
    if param:
      paramValues = self.getFilesParam(lfns, param)
      lfns = [f for f, v in paramValues.items() if v == paramValue]  # can be an iterator

    if lfns:
      lfnToCheck = lfns[0]
    else:
      return 0
    #
    # If files are RAW files, no need to get the number of ancestors!
    #
    if self.usingRAWFiles is None:
      self.usingRAWFiles = self.getMetadataFromTSorBK(lfnToCheck, 'FileType')[lfnToCheck] == 'RAW'
    if self.usingRAWFiles:
      return len(lfns) if not getFiles else lfns
    #
    # Get number of ancestors for known files
    cachedLfns = self.cachedLFNAncestors.get(runID, {})
    # If we cached a number, clean the cache
    if cachedLfns and True in set(isinstance(val, six.integer_types)
                                  for val in cachedLfns.values()):  # can be an iterator
      cachedLfns = {}
    setLfns = set(lfns)
    hitLfns = setLfns & set(cachedLfns)
    if hitLfns and not getFiles:
      self.logVerbose("Ancestors cache hit for run %d: %d files cached" %
                      (runID, len(hitLfns)))
      for lfn in hitLfns:
        ancestorFiles.update(cachedLfns[lfn])
      lfns = setLfns - hitLfns

    # If some files are unknown, get the ancestors from BK
    if lfns:
      res = self.getFileAncestors(list(lfns))
      if res['OK']:
        ancestorDict = res['Value']['Successful']
      else:
        self.logError("Error getting ancestors: %s" % res['Message'])
        ancestorDict = {}
      for lfn in ancestorDict:
        ancFiles = set(os.path.basename(f['FileName']) for f in ancestorDict[lfn] if f['FileType'] == 'RAW')
        ancestorFiles.update(ancFiles)
        self.cachedLFNAncestors.setdefault(runID, {})[lfn] = ancFiles

    if getFiles:
      return ancestorFiles
    notProcessed = self.__getNotProcessedAncestors(runID, lfnToCheck)
    ancestors = len(ancestorFiles)
    if notProcessed:
      self.logVerbose("Found %d files not processed for run %d" % (notProcessed, runID))
      ancestors += notProcessed
    self.logVerbose("found %d for run %d" % (ancestors, runID))
    return ancestors

  def __getNotProcessedAncestors(self, runID, lfnToCheck):
    """returns the number of RAW ancestor files that were not processed by the
    reconstruction production This is necessary only if that produciton is not
    processing all files, but in the doubt we check it."""
    if runID in self.notProcessed or not lfnToCheck:
      return self.notProcessed.get(runID, 0)
    ancestorFullDST = None
    recoProduction = None
    notProcessed = 0
    if not self.__recoType:
      self.__recoType = self.getPluginParam('RecoFileType', ['FULL.DST'])
    # Check if the file itself is a FULL.DST
    res = self.bkClient.getFileMetadata(lfnToCheck)
    if res['OK']:
      if res['Value']['Successful'].get(lfnToCheck, {}).get('FileType') in self.__recoType:
        ancestorFullDST = lfnToCheck
      else:
        # If not, get its ancestors
        res = self.getFileAncestors(lfnToCheck, replica=False)
        if res['OK']:
          fullDst = [f['FileName'] for f in res['Value']['Successful'].get(
              lfnToCheck, [{}]) if f.get('FileType') in self.__recoType]
          if fullDst:
            ancestorFullDST = fullDst[0]
        else:
          self.logError("Error getting ancestors of %s" % lfnToCheck, res['Message'])
    if ancestorFullDST:
      self.logDebug("Ancestor %s found: %s" % (self.__recoType, ancestorFullDST))
      res = self.bkClient.getJobInfo(ancestorFullDST)
      if res['OK']:
        try:
          recoProduction = res['Value'][0][18]
          self.logVerbose('Reconstruction production is %d' % recoProduction)
        except IndexError as e:
          self.logException("Exception extracting reco production from %s" % str(res['Value']), lException=e)
          recoProduction = None
      else:
        self.logError("Error getting job information", res['Message'])
    else:
      self.logVerbose("No ancestor %s file found" % self.__recoType)
    if recoProduction:
      res = self.transClient.getTransformationFiles({'TransformationID': recoProduction, 'RunNumber': runID})
      if res['OK']:
        notProcessed = len([fileDict for fileDict in res['Value'] if fileDict['Status'] == 'NotProcessed'])
    self.notProcessed[runID] = notProcessed
    return notProcessed

  # @timeThis
  def getTransformationFiles(self, runID):
    return self.transClient.getTransformationFiles({'TransformationID': self.transID, 'RunNumber': runID})

  # @timeThis
  def getFileAncestors(self, lfns, depth=10, replica=True):
    return self.bkClient.getFileAncestors(lfns, depth=depth, replica=replica)

  def checkAncestorsAtSE(self, taskLfns, seList):
    """Check if ancestors of a list of files are present at a set of SEs.

    :param list taskLfns: list of lfn lists for all tasks
    :param list seList: list of SEs to consider

    :return int missingAtSEs: number of files that are not at the specified SEs
    """
    # We request also that ancestors are at the fromSEs
    lfns = [lfn for lfnList in taskLfns for lfn in lfnList]
    nbLfns = len(lfns)
    ancestors = self.getFileAncestors(lfns, depth=1, replica=True)
    if not ancestors['OK']:
      self.logError("Error getting ancestors", ancestors['Message'])
      return _clearTaskLFNs(taskLfns)
    ancestors = ancestors['Value']['Successful']
    ancLfns = [anc['FileName'] for ancList in ancestors.values() for anc in ancList]  # can be an iterator
    self.logVerbose('Checking ancestors presence at %s for %d files' % (','.join(sorted(seList)),
                                                                        len(ancLfns)))
    res = self.dm.getReplicasForJobs(ancLfns, getUrl=False)
    if not res['OK']:
      self.logError("Error getting replicas of ancestors", res['Message'])
      # Fake no replicas for all files
      return _clearTaskLFNs(taskLfns)
    else:
      success = res['Value']['Successful']
    for lfn, ancList in ancestors.items():  # can be an iterator
      for anc in ancList:
        if not seList & set(success[anc['FileName']]):
          # Remove the LFN from the list of lists
          for lfnList in taskLfns:
            if lfn in lfnList:
              lfnList.remove(lfn)
    missingAtSEs = nbLfns - len([lfn for lfnList in taskLfns for lfn in lfnList])
    if missingAtSEs:
      self.logVerbose("%d ancestor files found not to be at %s" % (missingAtSEs, ','.join(sorted(seList))))
    return missingAtSEs

  # @timeThis
  def getTransformationRuns(self, runs=None, transID=None):
    """get the run table for a list of runs, if missing, add them.

    :param (list,tuple,dict,set,int,long): iterable with run numbers or single run number
    :param int transID: transformation ID. If not speficied, use the current ID

    :return S_OK / S_ERROR: value is the list of run dictionaries
    """
    if transID is None:
      transID = self.transID
    condDict = {'TransformationID': transID}
    if isinstance(runs, (dict, list, tuple, set)):
      condDict['RunNumber'] = list(runs)
    elif isinstance(runs, six.integer_types):
      runs = [runs]
      condDict['RunNumber'] = runs
    result = self.transClient.getTransformationRuns(condDict)
    if not result['OK']:
      return result
    # Check that all runs are there, if not add them
    runsFound = set(run['RunNumber'] for run in result['Value'])
    if runs:
      missingRuns = set(runs) - runsFound
      if missingRuns:
        self.logInfo('Add missing runs in transformation runs table: %s' %
                     ','.join(str(run) for run in sorted(missingRuns)))
        for runID in missingRuns:
          res = self.transClient.insertTransformationRun(transID, runID, '')
          if not res['OK']:
            return res
        result = self.transClient.getTransformationRuns(condDict)
    return result

  # @timeThis
  def getFilesGroupedByRunAndParam(self, lfns=None, param=''):
    """Group files by run and another BK parameter (e.g. file type or event
    type)"""
    runDict = {}
    # no need to query the BK as we have the answer from files
    if lfns is None:
      files = self.transFiles
    else:
      files = [fileDict for fileDict in self.transFiles if fileDict['LFN'] in lfns]
    self.logVerbose("Starting getFilesGroupedByRunAndParam for %d files, %s" %
                    (len(files), 'by %s' % param if param else 'no param'))
    runGroups = groupByRun(files)
    for runNumber, runLFNs in runGroups.items():  # can be an iterator
      if not param:
        runDict[runNumber] = {None: runLFNs}
      else:
        runDict[runNumber] = {}
        resDict = self.getFilesParam(runLFNs, param)
        for lfn, paramValue in resDict.items():  # can be an iterator
          runDict[runNumber].setdefault(paramValue, []).append(lfn)

    # If necessary fix files with run number 0
    zeroRunDict = runDict.pop(0, None)
    if zeroRunDict:
      nZero = 0
      for paramValue, zeroRun in zeroRunDict.items():  # can be an iterator
        newRuns = self.setRunForFiles(zeroRun)
        for newRun, runLFNs in newRuns.items():  # can be an iterator
          runDict.setdefault(newRun, {}).setdefault(paramValue, []).extend(runLFNs)
          nZero += len(runLFNs)
      self.logInfo("Set run number for %d files with run #0, which means it was not set yet" % nZero)

    if param:
      self.logVerbose("Grouped %d files by run and %s" % (len(files), param))
    return runDict

  def getFilesGroupedByRun(self, lfns=None):
    """Get files per run in a dictionary and set the run number if not done."""
    # Split the files in run groups
    if lfns is None:
      files = self.transFiles
    else:
      files = [fileDict for fileDict in self.transFiles if fileDict['LFN'] in lfns]
    runGroups = groupByRun(files)
    if not runGroups:
      # No files, no tasks!
      self.logVerbose("No runs found!")
      return []

    # If some files don't have a run number, get it and set it
    zeroRun = runGroups.pop(0, None)
    if zeroRun:
      nZero = 0
      newRuns = self.setRunForFiles(zeroRun)
      for newRun, runLFNs in newRuns.items():  # can be an iterator
        runGroups.setdefault(newRun, []).extend(runLFNs)
        nZero += len(runLFNs)
      self.logInfo("Set run number for %d files with run #0, which means it was not set yet" % nZero)
    self.logVerbose('Obtained %d runs' % len(runGroups))
    return runGroups

  def createTasks(self, storageElementGroups, chunkSize=None):
    """Create reasonable size tasks."""
    tasks = []
    if not chunkSize:
      chunkSize = self.getPluginParam('MaxFilesPerTask', 100)
    for stringTargetSEs in sorted(storageElementGroups):
      stringTargetLFNs = storageElementGroups[stringTargetSEs]
      for lfnGroup in breakListIntoChunks(sorted(stringTargetLFNs), chunkSize):
        tasks.append((stringTargetSEs, lfnGroup))
    self.logVerbose("%d tasks created" % len(tasks))
    return tasks

  def readCacheFile(self, workDirectory):
    """Utility function."""
    import pickle
    # Now try and get the cached information
    tmpDir = os.environ.get('TMPDIR', '/tmp')
    cacheFiles = ((workDirectory, ('TransPluginCache')),
                  (tmpDir, ('dirac', 'TransPluginCache')))
    for (cacheFile, prefixes) in cacheFiles:
      if not cacheFile:
        continue
      if isinstance(prefixes, str):
        prefixes = [prefixes]
      for node in prefixes:
        cacheFile = os.path.join(cacheFile, node)
        mkDir(cacheFile)
      cacheFile = os.path.join(cacheFile, "Transformation_%s.pkl" % (str(self.transID)))
      if not self.cacheFile:
        self.cacheFile = cacheFile
      try:
        with open(cacheFile, 'r') as cacheFile_o:
          self.cachedLFNAncestors = pickle.load(cacheFile_o)
          # Do not cache between cycles, only cache temporarily, but keep same structure in file, i.e. fake load
          _cachedNbRAWFiles = pickle.load(cacheFile_o)
          _cachedLFNSize = pickle.load(cacheFile_o)
          self.cachedRunLfns = pickle.load(cacheFile_o)
          self.cachedProductions = pickle.load(cacheFile_o)
          self.cachedLastRun = pickle.load(cacheFile_o)
          self.cachedLFNProcessedPath = pickle.load(cacheFile_o)
          self.lastCall = pickle.load(cacheFile_o)
        self.logVerbose("Cache file %s successfully loaded" % cacheFile)
        # print '*****'
        # print '\n'.join( ['%s %s' % ( key, val ) for key, val in self.cachedLFNProcessedPath.items()] )
        break
      except IOError:
        self.logVerbose("Cache file %s could not be loaded" % cacheFile)

  def getCachedRunLFNs(self, runID, paramValue):
    """Keep track of all files for a given parameter value."""
    return set(self.cachedRunLfns.get(runID, {}).get(paramValue, []))

  def setCachedRunLfns(self, runID, paramValue, lfnList):
    """Cache the list of LFNs for a given parameter value."""
    self.cachedRunLfns.setdefault(runID, {})[paramValue] = lfnList

  def getCachedProductions(self):
    return self.cachedProductions

  def setCachedProductions(self, productions):
    self.cachedProductions = productions

  def getCachedLastRun(self):
    return self.cachedLastRun

  def setCachedLastRun(self, lastRun):
    self.cachedLastRun = lastRun

  def cacheExpired(self, runID):
    if runID not in self.__runExpired:
      self.__runExpired[runID] = (random.uniform(0., 1.) > self.cacheHitFrequency)
    return self.__runExpired[runID]

  # @timeThis
  def getNbRAWInRun(self, runID, evtType):
    """Get the number of RAW files in a run."""
    # Every now and then refresh the cache
    rawFiles = self.cachedNbRAWFiles.get(runID, {}).get(evtType)
    if not rawFiles:
      res = self.bkClient.getNbOfRawFiles({'RunNumber': runID, 'EventTypeId': evtType, 'Finished': 'Y'})
      if not res['OK']:
        self.logError("Cannot get number of RAW files for run %d, evttype %d" % (runID, evtType))
        return 0
      rawFiles = res['Value']
      self.cachedNbRAWFiles.setdefault(runID, {})[evtType] = rawFiles
      if rawFiles:
        self.logVerbose("Run %d has %d RAW files" % (runID, rawFiles))
      else:
        self.logVerbose("Run %d is not finished yet" % runID)
    return rawFiles

  def writeCacheFile(self):
    """Utility function."""
    import pickle
    if self.cacheFile:
      try:
        with open(self.cacheFile, 'w') as cacheFile_o:
          pickle.dump(self.cachedLFNAncestors, cacheFile_o)
          pickle.dump(self.cachedNbRAWFiles, cacheFile_o)
          pickle.dump(self.cachedLFNSize, cacheFile_o)
          pickle.dump(self.cachedRunLfns, cacheFile_o)
          pickle.dump(self.cachedProductions, cacheFile_o)
          pickle.dump(self.cachedLastRun, cacheFile_o)
          pickle.dump(self.cachedLFNProcessedPath, cacheFile_o)
          pickle.dump(self.lastCall, cacheFile_o)
        self.logVerbose("Cache file %s successfully written" % self.cacheFile)
      except Exception as e:
        self.logException("Could not write cache file", self.cacheFile, lException=e)

  def setRunForFiles(self, lfns):
    """For files that are missing the run number, set it from BK."""
    res = self.bkClient.getFileMetadata(lfns)
    runFiles = {}
    if res['OK']:
      for lfn, metadata in res['Value']['Successful'].items():  # can be an iterator
        runFiles.setdefault(metadata['RunNumber'], []).append(lfn)
      for run in sorted(runFiles):
        if not run:
          self.logInfo("%d files found with void run '%s': \n%s" % (
              len(runFiles[run]), str(run), '\n'.join(runFiles[run])))
          runFiles.pop(run)
          continue
        res = self.transClient.addTransformationRunFiles(self.transID, run, runFiles[run])
        # print run, runFiles[run], res
        if not res['OK']:
          self.logError("Error setting files to run", " - %d files to run %d: %s" %
                        (len(runFiles[run]), run, res['Message']))
          runFiles.pop(run)
    else:
      self.logError("Error getting metadata for %d files" % len(lfns), res['Message'])
    return runFiles

  def getTransQuery(self, transReplicas):
    """Get BK query for the current transformation."""
    res = self.transClient.getBookkeepingQuery(self.transID)
    if not res['OK']:
      self.logError("Failed to get BK query for transformation", res['Message'])
      transQuery = {}
    else:
      transQuery = res['Value']
    if 'ProcessingPass' not in transQuery:
      # Get processing pass of the first file... This assumes all have the same...
      lfn = list(transReplicas)[0]
      res = self.bkClient.getDirectoryMetadata(lfn)
      if not res['OK']:
        self.logError("Error getting directory metadata", res['Message'])
        return None
      transQuery = res['Value']['Successful'].get(lfn, [{}])[0]
      # Strip off most of it
      for key in ('FileType', 'Production'):
        transQuery.pop(key, None)
    return transQuery

  def cleanFiles(self, transFiles, transReplicas, status=None):
    """Remove from transFiles all files without a replica and set their
    status."""
    noReplicaFiles = []
    for fileDict in [fileDict for fileDict in transFiles]:
      if fileDict['LFN'] not in transReplicas:
        noReplicaFiles.append(fileDict['LFN'])
        transFiles.remove(fileDict)
    if noReplicaFiles:
      if status is None:
        status = 'Problematic' if self.plugin not in getRemovalPlugins() else 'Processed'
      info = 'without replicas' if status == 'Problematic' else 'already processed'
      res = self.transClient.setFileStatusForTransformation(self.transID, status, noReplicaFiles)
      if not res['OK']:
        self.logError('Error setting file status for %d files' % len(noReplicaFiles), res['Message'])
      else:
        self.logInfo('Found %d files %s, status set to %s' % (len(noReplicaFiles), info, status))

  def __getRunDestinations(self, runIDList):
    """Get the site destination for a set of runs and cache this information in
    memory."""
    runSet = set(runIDList) - set(self.runDestinations)
    if runSet:
      # Try and get a run destination from TS
      res = self.transClient.getDestinationForRun(list(runSet))
      if res['OK']:
        dest = res['Value']
        if dest:
          if isinstance(dest, list):
            dest = dict(dest)
          for runID in [runID for runID in runSet if dest.get(runID)]:
            self.runDestinations[runID] = dest[runID]

  def getSEForDestination(self, runID, targets):
    """for a given run, get the information on destination SE from within a
    list of SEs."""
    self.__getRunDestinations([runID])
    site = self.runDestinations.get(runID)
    if site:
      self.logVerbose('Destination found for run %d: %s' % (runID, site))
      # Avoid spurious warning if no match found
      savedLevel = gLogger.getLevel()
      gLogger.setLevel('ERROR')
      res = self.dmsHelper.getSEInGroupAtSite(targets, site)
      gLogger.setLevel(savedLevel)
      return res.get('Value')
    return None

  def _selectProcessedFiles(self, lfns, prodList):
    """
    Select only the files that are not in a status other than Processed in a list of productions
    Beware: this is not the same as Processed files as some files may not be in any production
            because they are intermediate files
    """
    startTime = time.time()
    res = self.transClient.getTransformationFiles({'LFN': list(lfns), 'TransformationID': prodList})
    if not res['OK']:
      self.logError("Error getting transformation files", res['Message'])
      return {}
    fDictList = res['Value']
    unProcessedLfns = set(fDict['LFN'] for fDict in fDictList if fDict['Status'] != 'Processed')
    procLfns = lfns - unProcessedLfns
    prodLfns = {}
    processedLfns = set()
    for fDict in fDictList:
      lfn = fDict['LFN']
      if lfn in procLfns:
        processedLfns.add(lfn)
        prodLfns.setdefault(fDict['TransformationID'], []).append(lfn)
    # This is for files that are not in any production, accept them by default
    notInProdLfns = procLfns - processedLfns
    if notInProdLfns:
      processedLfns = procLfns
      for prod in prodList:
        prodLfns.setdefault(prod, []).extend(list(notInProdLfns))

    if len(processedLfns) != len(lfns):
      self.logVerbose('Reduce files from %d to %d '
                      '(removing non-processed files) in %.1f s' % (len(lfns),
                                                                    len(processedLfns),
                                                                    time.time() - startTime))
    return prodLfns

  def __getChunkSize(self):
    """Cache the chunk size."""
    if self.chunkForFileDescendants is None:
      self.chunkForFileDescendants = self.getPluginParam('MaxFilesToGetDescendants', 100)
    return self.chunkForFileDescendants

  def checkForDescendants(self, lfnSet, prodList, depth=0):
    """check if the files have an existing descendant in the list of
    productions."""
    finalResult = set()
    if not lfnSet or not prodList:
      return S_OK(finalResult)
    self.logVerbose("Checking descendants for %d files in productions %s, depth %d" %
                    (len(lfnSet),
                     ','.join(str(prod) for prod in prodList),
                     depth))
    excludeTypes = ('LOG', 'BRUNELHIST', 'DAVINCIHIST', 'GAUSSHIST', 'HIST', 'INDEX.ROOT')
    if isinstance(lfnSet, six.string_types):
      lfnSet = {lfnSet}
    elif isinstance(lfnSet, list):
      lfnSet = set(lfnSet)
    # Select files that have been processed in one of the productions,
    #   and return the list of these prods
    prodLfns = self._selectProcessedFiles(lfnSet, prodList)
    if not prodLfns:
      return S_OK(finalResult)
    # Get daughters of processed files in each production
    success = {}
    descProd = {}
    chunkSize = self.__getChunkSize()
    for prod, lfns in prodLfns.items():  # can be an iterator
      progressBar = ProgressBar(len(lfns),
                                title="Getting descendants for %d files in production %d" %
                                (len(lfns), prod), chunk=chunkSize,
                                interactive=sys.stdout.isatty(), log=self.logVerbose)
      for lfnChunk in breakListIntoChunks(lfns, chunkSize):
        progressBar.loop()
        # prod is a long and server expects an int!
        res = self.bkClient.getFileDescendants(lfnChunk, depth=1, production=int(prod), checkreplica=False)
        if not res['OK']:
          return res
        for lfn, descDict in res['Value']['WithMetadata'].items():  # can be an iterator
          # Only keep the file type as metadata in teh dict
          success.setdefault(lfn, {}).update(dict((desc, {'FileType': metadata['FileType'],
                                                          'GotReplica': metadata['GotReplica']})
                                                  for desc, metadata in descDict.items()  # can be an iterator
                                                  if metadata['FileType'] not in excludeTypes))
          # Record information about which production created each descendant
          for desc in descDict:
            descProd[desc] = prod
      progressBar.endLoop()

    # Get set of file types
    fileTypes = sorted(set(metadata['FileType']
                           for descendants in success.values()
                           for metadata in descendants.values()))  # can be an iterator
    self.logVerbose("Will check file type%s %s" % ('s' if len(fileTypes) > 1 else '', ','.join(fileTypes)))

    # Try and find descendants for each file type in turn as this is sufficient
    for fileType in fileTypes:
      # Invert list of descendants: descToCheck has descendant as key and list of parents
      descToCheck = {}
      descMetadata = {}
      for lfn, descendants in success.items():  # can be an iterator
        if lfn not in finalResult:
          descMetadata.update(descendants)
          for desc, metadata in descendants.items():  # can be an iterator
            if metadata['FileType'] == fileType:
              descToCheck.setdefault(desc, set()).add(lfn)
      if not descToCheck:
        break

      prodDesc = {}
      # Check if we found a descendant in the requested productions
      foundProds = {}
      for desc in list(descToCheck):  # pylint: disable=consider-iterating-dictionary
        prod = descProd[desc]
        # If we found an existing descendant in the list of productions, all OK
        if descMetadata[desc]['GotReplica'] == 'Yes' and prod in prodList:
          foundProds.setdefault(prod, set()).update(descToCheck[desc])
          finalResult.update(descToCheck[desc])
          del descToCheck[desc]
        else:
          # Sort remaining descendants in productions
          prodDesc.setdefault(prod, set()).add(desc)
      for prod in foundProds:
        self.logVerbose("Found %s descendants with replicas for %d files (out of %d) in production %d at depth %d" %
                        (fileType, len(foundProds[prod]), len(lfnSet), prod, depth + 1))

      # Check descendants without replicas in reverse order of productions
      for prod in sorted(prodDesc, reverse=True):
        # Check descendants whose parent was not yet found processed
        toCheckInProd = set(desc for desc in prodDesc[prod] if not descToCheck[desc] & finalResult)
        if toCheckInProd:
          self.logVerbose("Found %d %s descendants without replicas in production %d at depth %d,"
                          " check them for descendants" %
                          (len(toCheckInProd), fileType, prod, depth + 1))
          res = self.checkForDescendants(toCheckInProd, prodList, depth=depth + 1)
          if not res['OK']:
            return res
          foundInProd = set()
          for desc in res['Value']:
            foundInProd.update(descToCheck[desc])
          if foundInProd:
            self.logVerbose("Found %s descendants of %d files at depth %d" % (fileType, len(foundInProd), depth + 2))
            finalResult.update(foundInProd)

    return S_OK(finalResult)

  def __getStorageUsage(self, dirList, seList):
    """Get from StorageUsage the actual number of files for a list of
    directories at a list of SEs."""
    suClient = StorageUsageClient()
    result = {}
    if isinstance(seList, (dict, set, tuple)):
      seList = list(seList)
    self.logVerbose("Get storage usage for directories", ','.join(dirList))
    for dirName in dirList:
      res = suClient.getStorageSummary(dirName, None, None, seList)
      if not res['OK']:
        return res
      for se, stat in res['Value'].items():  # can be an iterator
        result[se] = result.setdefault(se, 0) + stat['Files']
    return S_OK(result)

  def _countRecentFiles(self, recentFiles, status, destSEs, dirList, delay):
    """Get files in a given status, and count how many are targeted for SEs in
    a list with a lastUpdate more recent than 'delay' hours."""
    self.logDebug("Getting %s files in dir %s at %s in last %d hours" %
                  (status, ','.join(dirList), ','.join(destSEs), delay))
    condDict = {'TransformationID': self.transID, 'Status': status}
    res = self.transClient.getTransformationFiles(condDict)
    if not res['OK']:
      return res
    limitTime = datetime.datetime.now() - datetime.timedelta(hours=delay)
    for fileDict in res['Value']:
      dirName = stripDirectory(fileDict['LFN'])
      usedSE = fileDict['UsedSE']
      if dirName in dirList and usedSE in destSEs and (not delay or fileDict['LastUpdate'] > limitTime):
        recentFiles[usedSE] += 1
    self.printShares("Number of files %s in transformation %s (%d hours):" %
                     (status, condDict['TransformationID'], delay),
                     recentFiles, counters=[], log=self.logDebug)
    return S_OK()

  def __getMaxFilesAtSE(self, targetFilesAtDestination, directories, destSEs):
    """Get the number of files already present at SEs for a list of LFN
    directories Using the processing and RAW distribution shares, split the
    maximum number of files to be staged on these SEs and return the number of
    files that can possibly be added."""
    # We strip off the directory to a default depth of 4 as we ignore the production
    dirList = stripDirectory(directories)
    # Add possibility to throttle the frequency of the plugin, but not clear if this is useful
    period = self.getPluginParam('Period', 0)
    now = datetime.datetime.utcnow()
    if self.lastCall and (now - self.lastCall) < datetime.timedelta(hours=period):
      self.logInfo("Skip this loop (less than %s hours since last call)" % period)
      return S_OK(None)
    self.lastCall = now
    self.writeCacheFile()
    self.logVerbose("Checking how many files can be sent to %s\n\t for directories %s" %
                    (','.join(destSEs), ','.join(dirList)))

    # Get current storage usage per SE
    storageUsage = self.__getStorageUsage(dirList, destSEs)
    if not storageUsage['OK']:
      return storageUsage
    storageUsage = storageUsage['Value']
    self.printShares("Current storage usage per SE:", storageUsage, counters=[], log=self.logVerbose)

    # Get the shares (including processing fraction) at each site (from RAW SEs)
    rawShares = self.getPluginShares(section='CPUforRAW', backupSE='CERN-RAW', log=self.logDebug, counters=False)
    if not rawShares['OK']:
      return rawShares
    # Turn a RAW share into a share on selected SEs
    shares = {}
    for rawSE, share in rawShares['Value'][1].items():  # can be an iterator
      selectedSEs = self.closerSEs([rawSE], destSEs, local=True)
      if selectedSEs:
        share *= targetFilesAtDestination / 100.
        shares[selectedSEs[0]] = share

    self.printShares("Shares per SE for %d files:" % targetFilesAtDestination, shares,
                     counters=[], log=self.logVerbose)

    recentFiles = dict.fromkeys(destSEs, 0)
    # Get the number of files already assigned for each SE, as it should be substracted from possible number
    res = self._countRecentFiles(recentFiles, 'Assigned', destSEs, dirList, 0)
    if not res['OK']:
      return res

    # Get the number of files Processed in the last 12 hours
    delay = 12
    res = self._countRecentFiles(recentFiles, 'Processed', destSEs, dirList, delay)
    if not res['OK']:
      return res
    self.printShares("Number of files Assigned or recently Processed (%d hours):" % delay,
                     recentFiles, counters=[], log=self.logVerbose)

    # Share targetFilesAtDestination on the SEs taking into account current usage
    maxFilesAtSE = {}
    for se, share in shares.items():  # can be an iterator
      maxFilesAtSE[se] = max(0, int(share - storageUsage.get(se, 0) - recentFiles.get(se, 0)))
    self.printShares("Maximum number of files per SE:", maxFilesAtSE, counters=[], log=self.logInfo)
    return S_OK(maxFilesAtSE)

  def getMaxFilesToReplicate(self, workDirectory):
    """Get the watermark and max files per SE if defined The keys of
    maxFilesAtSE are the destination SEs If plugin should just return, set
    watermark to None."""
    destSEs = set(resolveSEGroup(self.getPluginParam('DestinationSEs', [])))
    maxFilesAtSE = dict.fromkeys(destSEs, sys.maxsize)
    if not destSEs:
      self.logWarn('No destination SE given')
      return S_OK((None, maxFilesAtSE))
    # Check if destination is tape; no throttling is required
    allTapeSEs = all(StorageElement(se).status()['TapeSE'] for se in destSEs)
    if allTapeSEs:
      self.logVerbose("All SEs are on tape: no throttling")
      return S_OK((0, maxFilesAtSE))
    watermark = self.getPluginParam('MinFreeSpace', 30)
    targetFilesAtDestination = self.getPluginParam('TargetFilesAtDestination', 0)

    # if there is a maximum number of files to get at destination, get the current usage
    if targetFilesAtDestination:
      self.readCacheFile(workDirectory)
      # Directories limited to the top 4 directories
      directories = stripDirectory(self.transReplicas)
      # Get the maximum number of files that are allowed to be copied at this round (for prestaging mainly)
      result = self.__getMaxFilesAtSE(targetFilesAtDestination, directories, destSEs)
      if not result['OK']:
        return result
      # This happens when the cycle is to be skipped, then return None
      if not result['Value']:
        watermark = None
      # Update dictionary
      maxFilesAtSE.update(result['Value'])
    return S_OK((watermark, maxFilesAtSE))

  def _getFileSize(self, lfns):
    """Overwrite the DIRAC method and get the file size from the TS tables."""
    return S_OK(self.getMetadataFromTSorBK(lfns, 'FileSize'))

  def checkCondDBRunTick(self, runID):
    """Check for the presence of the run tick in the ConDB."""
    if not self.onlineCondDB:
      self.onlineCondDB = self.getPluginParam("OnlineCondDB", "/cvmfs/lhcb.cern.ch/lib/lhcb/git-conddb/ONLINE.git")

    if not os.path.exists(self.onlineCondDB):
      self.logWarn("OnlineCondDB not found", self.onlineCondDB)
      return None
    cmd = "git --git-dir %s  grep -q '^%d$' HEAD -- Conditions/Online/valid_runs.txt" % (self.onlineCondDB, int(runID))
    retCode = os.system(cmd)
    return not bool(retCode)

# Set of utility functions used by LHCbDirac transformation system


def getRemovalPlugins():
  """Returns list of removal plugins."""
  return ("DestroyDataset", 'DestroyDatasetWhenProcessed', "RemoveDatasetFromDisk",
          'RemoveReplicasKeepDestination', "ReduceReplicasKeepDestination",
          "RemoveReplicas", 'RemoveReplicasWhenProcessed',
          'RemoveReplicasWithAncestors', 'ReduceReplicas',)


def getReplicationPlugins():
  """Returns list of replication plugins."""
  return ("LHCbDSTBroadcast", "LHCbMCDSTBroadcastRandom", "LHCbWGBroadcastRandom",
          "ArchiveDataset", "ReplicateDataset",
          'RAWReplication', "ReplicateToRunDestination",
          'FakeReplication', 'ReplicateToLocalSE', 'ReplicateWithAncestors',
          'Healing',)


def getShares(sType, normalise=False):
  """Get the shares from the Operations section of the CS If not found, look in
  the Resources section."""
  optionPath = 'Shares/%s' % sType
  res = Operations().getOptionsDict(optionPath)
  if not res['OK']:
    res = gConfig.getOptionsDict(os.path.join('/Resources', optionPath))
  if not res['OK']:
    return res
  if not res['Value']:
    return S_ERROR("/Resources/%s option contains no shares" % optionPath)
  shares = res['Value']
  for site, value in shares.items():  # can be an iterator
    shares[site] = float(value)
  if normalise:
    shares = normaliseShares(shares)
  if not shares:
    return S_ERROR("No non-zero shares defined")
  return S_OK(shares)


def normaliseShares(shares):
  """Normalise to 1 the sum of shares.

  :param shares: shares per SE or site
  :type shares: dict
  """
  total = 0.0
  normShares = shares.copy()
  for site in shares:
    total += float(shares[site])
  if not total:
    return shares
  for site in shares:
    normShares[site] = 100.0 * (float(shares[site]) / total)
  return normShares

# @timeThis


def groupByRun(files):
  """Groups files by run files is a list of dictionaries containing the run
  number."""
  runDict = {}
  for fileDict in files:
    runID = fileDict.get('RunNumber')
    lfn = fileDict['LFN']
    if lfn:
      runDict.setdefault(runID if runID else 0, []).append(lfn)
  return runDict


def addFilesToTransformation(transID, lfns, addRunInfo=True):
  """Add files to a transformation, including the run number if required As
  this is also used by the add-files script, we also add run metadata to the TS
  is not present."""
  transClient = TransformationClient()
  bk = BookkeepingClient()
  gLogger.info("Adding %d files to transformation %s" % (len(lfns), transID))
  res = transClient.getTransformation(transID)
  if not res['OK']:
    return res
  transPlugin = res['Value']['Plugin']
  pluginsWithRunInfo = Operations().getValue('TransformationPlugins/PluginsWithRunInfo', [])
  addRunInfo = addRunInfo and transPlugin in pluginsWithRunInfo
  addedLfns = set()
  for lfnChunk in breakListIntoChunks(lfns, 1000):
    runDict = {}
    if addRunInfo:
      res = bk.getFileMetadata(lfnChunk)
      if res['OK']:
        resMeta = res['Value']['Successful']
        for lfn, metadata in resMeta.items():  # can be an iterator
          runID = metadata.get('RunNumber')
          if runID:
            runDict.setdefault(int(runID), set()).add(lfn)
      else:
        break
    errorLogged = False
    while True:
      res = transClient.addFilesToTransformation(transID, lfnChunk)
      if not res['OK']:
        if not errorLogged:
          errorLogged = True
          gLogger.error("Error adding %d files to transformation, retry..." % len(lfnChunk), res['Message'])
        time.sleep(1)
      else:
        break
    added = set(lfn for (lfn, status) in res['Value']['Successful'].items() if status == 'Added')  # can be an iterator
    addedLfns.update(added)
    if addRunInfo and res['OK']:
      gLogger.info("Add information for %d runs to transformation %s" % (len(runDict), transID))
      for runID, runLfns in runDict.items():  # can be an iterator
        runLfns &= added
        if runLfns:
          res = transClient.addTransformationRunFiles(transID, runID, list(runLfns))
          if not res['OK']:
            break
      # Add run metadata if not present in TS
      res = transClient.getRunsMetadata(list(runDict))
      if res['OK']:
        missingRuns = []
        for runID, meta in res['Value'].items():  # can be an iterator
          if 'TCK' not in meta or 'CondDb' not in meta or 'DDDB' not in meta:
            missingRuns.append(runID)
      else:
        missingRuns = list(runDict)
      if missingRuns:
        res = bk.getRunInformation({'RunNumber': missingRuns, 'Fields': ['TCK', 'CondDb', 'DDDB']})
        if not res['OK']:
          gLogger.error("Error getting run information", res['Message'])
        else:
          for runID, meta in res['Value'].items():  # can be an iterator
            res = transClient.addRunsMetadata(runID, meta)
            if not res['OK']:
              gLogger.error("Error setting run metadata in TS", res['Message'])

  if not res['OK']:
    return res
  gLogger.info("%d files successfully added to transformation" % len(addedLfns))
  return S_OK(addedLfns)


def stripDirectory(files, depth=None):
  """Return set of directories for a list of LFNs, and directory for a single LFN."""
  if depth is None:
    depth = 4
  if isinstance(files, six.string_types):
    return os.path.sep + os.path.join(*files.split(os.path.sep)[:depth + 1])
  return set(os.path.sep + os.path.join(*lfn.split(os.path.sep)[:depth + 1]) for lfn in files)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.AccountingSystem.private.Plotters.StoragePlotter.

StoragePlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.Storage import Storage

__RCSID__ = "$Id$"

# FIXME: refactor _reportMethods
# FIXME: refactor _plotMethods


class StoragePlotter(BaseReporter):
  """StoragePlotter as extension of BaseReporter."""

  _typeName = "Storage"
  _typeKeyFields = [dF[0] for dF in Storage().definitionKeyFields]

  # .............................................................................
  # catalog Space

  _reportCatalogSpaceName = "LFN size"

  def _reportCatalogSpace(self, reportRequest):
    """Reports about LFN size and catalog space from the accounting.

    :param reportRequest: <dict>
      { 'grouping'       : 'Directory',
        'groupingFields' : ( '%s', [ 'Directory' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'Directory' : [ '/lhcb/data', '/lhcb/LHCb' ] }
      }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { '/lhcb/data' : { 1355616000L : 4.9353885242469104,
                                             1355702400L : 4.8438444870748203 },
                            '/lhcb/LHCb' : { 1355616000L : 3.93538852424691,
                                             1355702400L : 3.8438444870748198 }
                           },
        'data'          : { '/lhcb/data' : { 1355616000L : 4935388.5242469106,
                                             1355702400L : 4843844.4870748203 },
                            '/lhcb/LHCb' : { 1355616000L : 3935388.5242469101,
                                             1355702400L : 3843844.4870748199 }
                           },
        'unit'          : 'TB',
        'granularity'   : 86400
       }
    """

    if reportRequest['grouping'] == 'StorageElement':
      return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s)/SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'LogicalSize', 'entriesInBucket'
                                                          ]
                    )
    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "bytes")

    # 3rd value, maxValue is not used
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotCatalogSpace(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the LFN
    size and the catalog space.

    :param reportRequest: <dict>
       { 'grouping'       : 'Directory',
         'groupingFields' : ( '%s', [ 'Directory' ] ),
         'startTime'      : 1355663249.0,
         'endTime'        : 1355749690.0,
         'condDict'       : { 'Directory' : [ '/lhcb/data', '/lhcb/LHCb' ] }
       }
    :param plotInfo: <dict> ( output of _reportCatalogSpace )
       { 'graphDataDict' : { '/lhcb/data' : { 1355616000L : 4.9353885242469104,
                                              1355702400L : 4.8438444870748203 },
                             '/lhcb/LHCb' : { 1355616000L : 3.93538852424691,
                                              1355702400L : 3.8438444870748198 }
                            },
         'data'          : { '/lhcb/data' : { 1355616000L : 4935388.5242469106,
                                              1355702400L : 4843844.4870748203 },
                             '/lhcb/LHCb' : { 1355616000L : 3935388.5242469101,
                                              1355702400L : 3843844.4870748199 }
                            },
         'unit'          : 'TB',
         'granularity'   : 86400
        }
    :param filename: <str>
      '_plotCatalogSpace'

    returns S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "LFN space usage by %s" % reportRequest['grouping'],
                'starttime': startEpoch,
                'endtime': endEpoch,
                'span': granularity,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # catalog Files

  _reportCatalogFilesName = "LFN files"

  def _reportCatalogFiles(self, reportRequest):
    """Reports about the LFN files and the catalog files from the accounting.

    :param reportRequest: <dict>
      { 'grouping'       : 'Directory',
        'groupingFields' : ( '%s', [ 'Directory' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'Directory' : [ '/lhcb/data', '/lhcb/LHCb' ] }
       }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { '/lhcb/data' : { 1355616000L : 4935388.5242469106,
                                             1355702400L : 4843844.4870748203 },
                            '/lhcb/LHCb' : { 1355616000L : 3935388.5242469101,
                                             1355702400L : 3843844.4870748199 }
                           },
        'data'          : { '/lhcb/data' : { 1355616000L : 4935388524246.9102,
                                             1355702400L : 4843844487074.8203 },
                            '/lhcb/LHCb' : { 1355616000L : 3935388524246.9102,
                                             1355702400L : 3843844487074.8198 }
                           },
        'unit'          : 'Mfiles',
        'granularity'   : 86400
       }
    """

    if reportRequest['grouping'] == 'StorageElement':
      return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s)/SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'LogicalFiles', 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "files")

    # 3rd value, maxValue is not used
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotCatalogFiles(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the LFN
    files and the catalog files.

    :param reportRequest: <dict>
      { 'grouping'       : 'Directory',
        'groupingFields' : ( '%s', [ 'Directory' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'Directory' : [ '/lhcb/data', '/lhcb/LHCb' ] }
       }
    :param plotInfo: <dict> ( output of _reportCatalogFiles )
      { 'graphDataDict' : { '/lhcb/data' : { 1355616000L : 4935388.5242469106,
                                             1355702400L : 4843844.4870748203 },
                            '/lhcb/LHCb' : { 1355616000L : 3935388.5242469101,
                                             1355702400L : 3843844.4870748199 }
                          },
        'data'          : { '/lhcb/data' : { 1355616000L : 4935388524246.9102,
                                             1355702400L : 4843844487074.8203 },
                            '/lhcb/LHCb' : { 1355616000L : 3935388524246.9102,
                                            1355702400L : 3843844487074.8198 }
                                    },
        'unit'          : 'Mfiles',
        'granularity'   : 86400
       }
    :param filename: <str>
      '_plotCatalogFiles'

    returns S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """
    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "Number of LFNs by %s" % reportRequest['grouping'],
                'starttime': startEpoch,
                'endtime': endEpoch,
                'span': granularity,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # Physical Space

  _reportPhysicalSpaceName = "PFN size"

  def _reportPhysicalSpace(self, reportRequest):
    """Reports about the PFN size and the physical space from the accounting.

    :param reportRequest: <dict>
      { 'grouping'       : 'StorageElement',
        'groupingFields' : ( '%s', [ 'StorageElement' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'StorageElement' : [ 'CERN-ARCHIVE', 'CERN-DST' ] }
       }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { 'CERN-ARCHIVE' : { 1355616000L : 2.34455676781291,
                                               1355702400L : 2.5445567678129102 },
                            'CERN-DST'     : { 1355616000L : 0.34455676781290995,
                                               1355702400L : 0.54455676781290996 }
                           },
        'data'          : { 'CERN-ARCHIVE' : { 1355616000L : 2344556.76781291,
                                               1355702400L : 2544556.76781291 },
                            'CERN-DST'     : { 1355616000L : 344556.76781290997,
                                               1355702400L : 544556.76781291002 }
                           },
        'unit'          : 'TB',
        'granularity'   : 86400
       }
    """

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'PhysicalSize', 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "bytes")

    # 3rd value, maxValue is not used
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({
        'data': baseDataDict,
        'graphDataDict': graphDataDict,
        'granularity': granularity,
        'unit': unitName
    })

  def _plotPhysicalSpace(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the PFN
    size and the physical space.

    :param reportRequest: <dict>
      { 'grouping'       : 'StorageElement',
        'groupingFields' : ( '%s', [ 'StorageElement' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'StorageElement' : [ 'CERN-ARCHIVE', 'CERN-DST' ] }
      }
    :param plotInfo: <dict> ( output of _reportPhysicalSpace )
      { 'graphDataDict' : { 'CERN-ARCHIVE' : { 1355616000L : 2.34455676781291,
                                               1355702400L : 2.5445567678129102 },
                            'CERN-DST'     : { 1355616000L : 0.34455676781290995,
                                               1355702400L : 0.54455676781290996 }
                          },
        'data'          : { 'CERN-ARCHIVE' : { 1355616000L : 2344556.76781291,
                                               1355702400L : 2544556.76781291 },
                            'CERN-DST'     : { 1355616000L : 344556.76781290997,
                                               1355702400L : 544556.76781291002 }
                          },
        'unit'          : 'TB',
        'granularity'   : 86400
       }
    :param filename: <str>
      '_plotPhysicalSpace'

    returns S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "PFN space usage by %s" % reportRequest['grouping'],
                'starttime': startEpoch,
                'endtime': endEpoch,
                'span': granularity,
                'ylabel': plotInfo['unit']}
    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # physical Files

  _reportPhysicalFilesName = "PFN files"

  def _reportPhysicalFiles(self, reportRequest):
    """Reports about the PFN files and the physical files from the accounting.

    :param reportRequest: <dict>
       { 'grouping'       : 'StorageElement',
         'groupingFields' : ( '%s', [ 'StorageElement' ] ),
         'startTime'      : 1355663249.0,
         'endTime'        : 1355749690.0,
         'condDict'       : { 'StorageElement' : [ 'CERN-ARCHIVE', 'CERN-DST', 'CERN-BUFFER' ] }
        }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { 'CERN-BUFFER'  : { 1355616000L : 250.65890999999999,
                                               1355702400L : 261.65890999999999 },
                            'CERN-ARCHIVE' : { 1355616000L : 412.65890999999999,
                                               1355702400L : 413.65890999999999 },
                            'CERN-DST'     : { 1355616000L : 186.65890999999999,
                                               1355702400L : 187.65890999999999 }
                           },
        'data'          : { 'CERN-BUFFER'  : { 1355616000L : 250658.91,
                                               1355702400L : 261658.91 },
                            'CERN-ARCHIVE' : { 1355616000L : 412658.90999999997,
                                               1355702400L : 413658.90999999997 },
                            'CERN-DST'     : { 1355616000L : 186658.91,
                                               1355702400L : 187658.91 }
                           },
        'unit'          : 'kfiles',
        'granularity'   : 86400
       }
    """

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'PhysicalFiles', 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxVal, "files")

    # 3rd value, maxValue is not used
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotPhysicalFiles(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the PFN
    files and the physical files.

    :param reportRequest: <dict>
       { 'grouping'       : 'StorageElement',
         'groupingFields' : ( '%s', [ 'StorageElement' ] ),
         'startTime'      : 1355663249.0,
         'endTime'        : 1355749690.0,
         'condDict'       : { 'StorageElement' : [ 'CERN-ARCHIVE', 'CERN-DST', 'CERN-BUFFER' ] }
        }
    :param plotInfo: <dict> ( output of _reportPhysicalFiles )
      { 'graphDataDict' : { 'CERN-BUFFER'  : { 1355616000L : 250.65890999999999,
                                               1355702400L : 261.65890999999999 },
                            'CERN-ARCHIVE' : { 1355616000L : 412.65890999999999,
                                               1355702400L : 413.65890999999999 },
                            'CERN-DST'     : { 1355616000L : 186.65890999999999,
                                               1355702400L : 187.65890999999999 }
                          },
        'data'          : { 'CERN-BUFFER'  : { 1355616000L : 250658.91,
                                               1355702400L : 261658.91 },
                            'CERN-ARCHIVE' : { 1355616000L : 412658.90999999997,
                                               1355702400L : 413658.90999999997 },
                            'CERN-DST'     : { 1355616000L : 186658.91,
                                               1355702400L : 187658.91 }
                          },
        'unit'          : 'kfiles',
        'granularity'   : 86400
       }
    :param filename: <str>
      '_plotPhysicalFiles'

    return S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "Number of PFNs by %s" % reportRequest['grouping'],
                'starttime': startEpoch,
                'endtime': endEpoch,
                'span': granularity,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # PFN vs LFN File Multiplicity

  _reportPFNvsLFNFileMultiplicityName = "PFN/LFN file ratio"

  def _reportPFNvsLFNFileMultiplicity(self, reportRequest):

    logicalField = "LogicalFiles"
    physicalField = "PhysicalFiles"

    return self._multiplicityReport(reportRequest, logicalField, physicalField)

  def _plotPFNvsLFNFileMultiplicity(self, reportRequest, plotInfo, filename):

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "Ratio of PFN/LFN files by %s" % reportRequest['grouping'],
                'starttime': startEpoch,
                'endtime': endEpoch,
                'span': granularity,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # PFN vs LFN Size Multiplicity

  _reportPFNvsLFNSizeMultiplicityName = "PFN/LFN size ratio"

  def _reportPFNvsLFNSizeMultiplicity(self, reportRequest):

    logicalField = "LogicalSize"
    physicalField = "PhysicalSize"

    return self._multiplicityReport(reportRequest, logicalField, physicalField)

  def _plotPFNvsLFNSizeMultiplicity(self, reportRequest, plotInfo, filename):

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "Ratio of PFN/LFN space used by %s" % reportRequest['grouping'],
                'starttime': startEpoch,
                'endtime': endEpoch,
                'span': granularity,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # helper methods

  def _multiplicityReport(self, reportRequest, logicalField, physicalField):

    # Step 1 get the total LFNs for each bucket
    selectFields = ("%s, %s, %s, SUM(%s)/SUM(%s)",
                    ['Directory', 'startTime', 'bucketLength', logicalField, 'entriesInBucket']
                    )
    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                ('%s', ['Directory']),
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    # 2nd element ( granularity ) is unused
    dataDict, __ = retVal['Value']
    self.stripDataField(dataDict, 0)
    bucketTotals = self._getBucketTotals(dataDict)

    # Step 2 get the total PFNs
    _selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (_selectField + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          physicalField, 'entriesInBucket'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    # Step 3 divide the PFNs by the total amount of LFNs
    finalData = {}

    # FIXME: TO BE replaced by a faster implementation ( see below )
    for k in dataDict:
      for bt in dataDict[k]:
        if bt in bucketTotals:
          if k not in finalData:
            finalData[k] = {}
          finalData[k][bt] = dataDict[k][bt] / bucketTotals[bt]

#    for key, bucketTotal in dataDict.items():
#      for bt in bucketTotal.values():
#        if bt in bucketTotals:
#          if key not in finalData:
#            finalData[ key ] = {}
#          finalData[ key ][ bt ] = bucketTotal[ bt ] / bucketTotals[ bt ]

    return S_OK({'data': finalData,
                 'graphDataDict': finalData,
                 'granularity': granularity,
                 'unit': 'PFN / LFN'})



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Bookkeeping file system."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import cmd
import pydoc
import shlex
import argparse

from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb
from LHCbDIRAC.BookkeepingSystem.Client.LHCB_BKKDBClient import LHCB_BKKDBClient

__RCSID__ = "$Id$"


#############################################################################
class LHCbBookkeepingCLI(cmd.Cmd):
  """class."""
  #############################################################################

  def __init__(self):
    """constructor."""
    cmd.Cmd.__init__(self)
    self.prompt = "$[/]$"
    self.bk = LHCB_BKKDBClient()
    self.diracAPI = DiracLHCb()
    self.currentPath = '/'
    self.do_setDataQualityFlags('OK')
    self.saveParser = argparse.ArgumentParser(description="Save LFNS", prog='save')
    self.saveParser.add_argument('filename', type=str, help='file name')
    self.saveParser.add_argument("-f", "--format", help="txt or py")
    self.saveParser.add_argument("-n", "--num", help="number of files to be saved")
    self.saveParser.add_argument("-c", "--with-fileCatalog", help="save POOL XML catalog in a given site")
    self.sites = {}
    self.sites = DMSHelpers().getShortSiteNames(withStorage=False, tier=(0, 1))

  #############################################################################
  def addCurrentPath(self, path):
    """add a path."""
    if path[0] != '/' and len(self.currentPath) == 1:
      self.currentPath += path
    else:
      self.currentPath += '/' + path

  #############################################################################
  def __printPrompt(self, path=None):
    """prints the prompt."""
    if path is None:
      path = '/'
    self.prompt = '$[' + path + ']$'

  #############################################################################
  def __bklist(self, path):
    """list a path."""
    retVal = self.bk.list(path)
    return retVal

  def __bkListAll(self, path):
    """list the directory with metadata.

    This is equivalent to ls -a
    """
    if path == '':
      path = self.currentPath
    res = self.bk.list(path)
    for i in res:
      print(i['name'])
      for j in i:
        if j not in ['fullpath', 'selection', 'expandable', 'method', 'level', 'name']:
          print('   ', j, i[j])

  #############################################################################
  def __checkDirectory(self, path):
    """is empty directory."""
    res = self.bk.list(path)
    retValue = False
    if res:
      retValue = True
    return retValue

  #############################################################################
  def __rootDirectory(self):
    """root."""
    self.currentPath = '/'
    self.__printPrompt(self.currentPath)

  #############################################################################
  def __oneLevelback(self, logical=False):
    """cd .."""
    path = self.currentPath.split('/')
    if path[0] == '' and path[1] == '':
      if not logical:
        self.currentPath = '/'
        self.__printPrompt(self.currentPath)
      else:
        return '/'
    else:
      newpath = ''
      for i in range(len(path) - 1):
        if path[i] != '':
          newpath += '/' + path[i]
      if newpath == '':
        newpath = '/'
      if not logical:
        self.currentPath = newpath
        self.__printPrompt(self.currentPath)
      else:
        return newpath

  #############################################################################
  @staticmethod
  def help_ls():
    """provides help."""
    print("Usage: ls [OPTION]... [FILE]...")
    print("List information about the FILEs (the current directory by default).")
    print(" Available options: -a")
    print('Usage: ls or ls -a')
    print('ls -a [FILE]')
    print('Note: You can do paging using the | more')
    print('For example: ls | more')

  @staticmethod
  def help_cd():
    """help cd command."""
    print(" cd <dir>")
    print("cd ..")
    print("cd /")
    print("cd")

  #############################################################################
  def do_ls(self, path):
    """ls command."""
    paging = False
    if path.find('|') > -1:
      tmpPath = path.split('|')
      path = ''
      for i in range(len(tmpPath) - 1):
        path += tmpPath[i].strip()
      paging = True

    text = ''
    if len(path) > 0 and path[0] == '-':
      try:
        if path[1] != 'a':
          print("ls: invalid option -- {}".format(path[1]))
          print("Try `help ls' for more information.")
        else:
          path = path[2:]
          self.__bkListAll(path)
      except IndexError as e:
        print("Invalid oprion:", e)
    else:
      if path == '':
        path = self.currentPath
      res = self.__bklist(path)
      for i in sorted(res, key=lambda x: x["fullpath"]):
        if paging:
          text += i['name'] + '\n'
        else:
          print(i['name'])
    if paging:
      pydoc.ttypager(text)

  #############################################################################
  def do_list(self, path):
    """list commamd."""
    pass

  #############################################################################
  def do_save(self, command):
    """save command."""
    try:
      args = self.saveParser.parse_args(shlex.split(command))
    except argparse.ArgumentError as exc:
      print(exc)
    except SystemExit:
      self.saveParser.print_help()
      return
    if args.format is None:
      args.format = 'txt'

    retVal = self.bk.list(self.currentPath)
    files = None
    if retVal:
      if 'FileName' not in retVal[0]:
        print('No file found belong to {} bookkeeping path!'.format(self.currentPath))
        return
      else:
        files = retVal
    else:
      print(retVal)
      return
    if args.num is not None:
      try:
        nb = int(args.num)
      except ValueError as v:
        print(v)
        return
      nbOfFiles = nb
    else:
      nbOfFiles = len(files)

    lfns = {}
    for i in range(nbOfFiles):
      lfns[files[i]['FileName']] = files[i]

    if args.with_fileCatalog:
      site = args.with_fileCatalog
      lfnList = list(lfns)
      totalFiles = len(lfnList)
      ff = args.filename.split('.')
      catalog = ff[0] + '.xml'
      retVal = self.diracAPI.getInputDataCatalog(lfnList, site, catalog, True)
      nbofsuccsessful = 0
      if retVal['OK']:
        slist = {}
        faild = {}
        if 'Successful' in retVal['Value']:
          slist = retVal['Value']['Successful']
        if 'Failed' in retVal['Value']:
          faild = retVal['Value']['Failed']
        nbofsuccsessful = len(slist)
        nboffaild = len(faild)
        exist = {}
        for i in slist:
          exist[i] = lfns[i]

        self.bk.writeJobOptions(exist,
                                args.filename,
                                savedType=None,
                                catalog=catalog,
                                savePfn=slist)

        message = 'Total files:' + str(totalFiles) + '\n'
        if site is not None:
          if nbofsuccsessful:
            message += str(nbofsuccsessful) + ' found ' + site + '\n'
          if nboffaild:
            message += str(nboffaild) + ' not found ' + site
        print(message)
        return
    if args.format == 'txt':
      text = self.bk.writeJobOptions(lfns, '', args.format)
    elif args.format == 'py':
      text = self.bk.writeJobOptions(lfns, args.filename, args.format)

    with open(args.filename, 'w') as f:
      f.write(text)

  #############################################################################
  def do_cd(self, path):
    """cd command."""
    newpath = self.currentPath + '/' + path
    if path == '':
      self.currentPath = '/'
      self.__printPrompt(self.currentPath)
    elif path == '..':
      self.__oneLevelback()
    elif path == '/':
      self.__rootDirectory()
    elif path[0] == '/':
      if self.__checkDirectory(path):
        self.currentPath = path
        self.__printPrompt(self.currentPath)
      else:
        print('No such file or directory.')
    elif self.__checkDirectory(newpath):
      self.addCurrentPath(path)
      self.__printPrompt(self.currentPath)
    else:
      print('No such file or directory.')

  #############################################################################
  def do_pwd(self, path):
    """pwd command."""
    print(self.currentPath)

  #############################################################################
  def do_queries(self, command=''):
    """execute query."""
    retVal = self.bk.getPossibleParameters()
    print('The following bookkeeping query types are available:')
    for i in retVal:
      print(' '.rjust(10) + i)
    print("You can change the query types using the 'use' command")

  #############################################################################
  def do_use(self, command):
    """use command."""
    self.bk.setParameter(str(command))
    self.do_cd('/')

  #############################################################################
  @staticmethod
  def help_use():
    """help of use command."""
    print('Usage:')
    print('  use type'.rjust(10))
    print('Arguments:')
    print(' type: bookkeeping query type'.rjust(10))
    print("The 'type' can be found using the 'queries' command!")
    print("EXAMPE:")
    print(' '.rjust(10) + "use 'Event type'")

  #############################################################################
  @staticmethod
  def help_queries():
    """help of queries command."""
    print("This method shows the available query types!")
    print("Usage:")
    print("  queries")
    print("You can choose a query type using the 'use' command  ")
    print(" NOTE: the default query type is 'Configuration'")

  #############################################################################
  def do_advanceQuery(self, command=''):
    """advancedQuery command."""
    self.bk.setAdvancedQueries(True)
    self.do_cd('/')

  #############################################################################
  @staticmethod
  def help_advanceQuery():
    """help."""
    print("It allows to see more level of the Bookkeeping Tree")
    print("Usage:")
    print("   advanceQuery")

  #############################################################################
  def do_standardQuery(self, command=''):
    """command."""
    self.bk.setAdvancedQueries(False)
    self.do_cd('/')

  #############################################################################
  @staticmethod
  def help_standardQuery():
    """help."""
    print("This is used by default")
    print("It shows a reduced bookkeeping path.")
    print("Usage:")
    print("   standardQuery")

  #############################################################################
  def do_dataQuality(self, command=''):
    """command."""
    print('The following Data Quality flags are available in the bookkeeping!')
    retVal = self.bk.getAvailableDataQuality()
    if retVal['OK']:
      for i in retVal['Value']:
        print(' '.ljust(10) + i)
    else:
      print(retVal["Message"])
    print("To set the data quality flags you have to use 'setDataQualityFlags' command!")
    print("More information: 'help setDataQualityFlags'")

  #############################################################################
  @staticmethod
  def help_dataQuality():
    """help."""
    print('This command shows the available data quality flags.')
    print("Usage:")
    print("  dataQuality")
    print('To change the data quality flag use the setDataQualityFlags command')

  #############################################################################
  def do_setDataQualityFlags(self, command):
    """command."""
    qualities = command.split(' ')
    if len(qualities) > 0:
      dataquality = {}
      for i in qualities:
        dataquality[i] = True
      self.bk.setDataQualities(dataquality)
    else:
      print('ERROR: Please give a data quality flag!')

  #############################################################################
  def __moreInfoProcpass(self, command):
    """more information of a directory."""
    found = False
    retVal = self.bk.getProcessingPassSteps({'StepName': command})
    if retVal['OK']:
      proc = retVal['Value']
      print('{0} {1} step found in the bkk'.format(proc['TotalRecords'], command))
      for i in proc['Records']:
        print(' '.ljust(5) + i)
        for j in proc['Records'][i]:
          print(' '.ljust(10) + str(j[0]) + ':' + str(j[1]))
        found = True
    else:
      print('ERROR: ', retVal['Message'])
    return found

  #############################################################################
  def do_moreinfo(self, command=''):
    """more info command."""
    if command == '':
      previouspath = self.__oneLevelback(self.currentPath)
      values = self.__bklist(previouspath)
      name = self.currentPath.split('/')
      for i in values:
        if 'level' in i and i['level'] == 'FileTypes':
          path = self.currentPath
          retVal = self.bk.getLimitedFiles({'fullpath': str(path)}, ['nb'], -1, -1)
          if not retVal['OK']:
            print(retVal['Message'])
            return
          print('The selected dataset is:')
          for selected in retVal['Value']['Extras']['Selection']:
            print(''.ljust(5) + selected + ' ' + str(retVal['Value']['Extras']['Selection'][selected]))
          print('Statistics:')
          print(' '.ljust(5) + 'Number of files:' + str(retVal['Value']['TotalRecords']))
          for selected in retVal['Value']['Extras']['GlobalStatistics']:
            print(''.ljust(5) + selected + ' ' + str(retVal['Value']['Extras']['GlobalStatistics'][selected]))
          break
        if i['name'] == name[len(name) - 1]:
          for j in i:
            if j not in ['fullpath', 'selection', 'expandable', 'method', 'level', 'name']:
              print('   ', j, i[j])
        if 'level' in i and i['level'] == 'Processing Pass':
          found = self.__moreInfoProcpass(command)
          break
    else:
      values = self.__bklist(self.currentPath)
      found = False
      for i in values:
        if i['name'] == command:
          for j in i:
            if j not in ['fullpath', 'selection', 'expandable', 'method', 'level', 'name']:
              print('   ', j, i[j])
          found = True
        if 'level' in i and i['level'] == 'Processing Pass':
          found = self.__moreInfoProcpass(command)
          break

      if not found:
        print(" The '%s' does not found" % (command))

  def do_sites(self, command):
    print("T0/1 sites: %s" % ','.join(site for site in self.sites))

  @staticmethod
  def help_sites(self):
    """help."""
    print("it returns a list of T1 sites")

  #############################################################################
  @staticmethod
  def help_setDataQualityFlags():
    """help."""
    print('This command allows to use different data quality flags.')
    print("Usage:")
    print("  setDataQualityFlags flag1 [flag2 flag3, ... flagN]")
    print("Arguments:")
    print("  flag[1...N]:  Data quality flags.")
    print('For example:')
    print(' '.ljust(10) + 'setDataQualityFlags OK UNCHECKED')

  #############################################################################
  @staticmethod
  def help_EOF():
    """quit."""
    print("Quits the program")

  #############################################################################
  def do_EOF(self, line):
    """quit command."""
    sys.exit()

  #############################################################################
  def help_save(self):
    """help."""
    self.saveParser.print_help()

  #############################################################################
  @staticmethod
  def help_moreinfo():
    """help method."""
    print("Display the statistics of the selected data.")
    print("Usage:")
    print("  moreinfo")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DMScript is a class that creates default switches for DM scripts, decodes
them and sets flags The module also provides a function for printing pretty
results from DMS queries."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import six
import sys
import time
import tempfile

import DIRAC
from DIRAC import gLogger, gConfig
from DIRAC.Core.Base import Script
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

__RCSID__ = "$Id$"


def __printDictionary(dictionary, offset=0, shift=0, empty="Empty directory", depth=9999):
  """Dictionary pretty printing."""
  key_max = 0
  value_max = 0
  for key, value in dictionary.items():
    key_max = max(key_max, len(str(key)))
    value_max = max(value_max, len(str(value)))
  center = key_max + offset
  newOffset = offset + (shift if shift else key_max)
  for key in sorted(dictionary):
    value = dictionary[key]
    if isinstance(value, dict):
      if not depth:
        value = list(value)
      elif value != {}:
        gLogger.notice('%s%s : ' % (offset * ' ', key))
        __printDictionary(value, offset=newOffset, shift=shift, empty=empty, depth=depth - 1)
      elif key not in ('Failed', 'Successful'):
        gLogger.notice('%s%s : %s' % (offset * ' ', key, empty))
    if isinstance(value, (list, set)):
      if not value:
        gLogger.notice('%s%s : %s' % (offset * ' ', key, '[]'))
      else:
        gLogger.notice('%s%s : ' % (offset * ' ', key))
        for val in sorted(value):
          gLogger.notice('%s%s' % (newOffset * ' ', val))
    elif not isinstance(value, dict):
      # In case value contains \n, indent the lines
      toPrint = str(value).split('\n')
      gLogger.notice('%s : %s' % (str(key).rjust(center), toPrint.pop(0)))
      for line in toPrint:
        gLogger.notice((center + 3) * ' ' + line)


def printDMResult(result, shift=4, empty="Empty directory", script=None, depth=999, offset=0):
  """Printing results returned with 'Successful' and 'Failed' items."""
  if not script:
    script = Script.scriptName
  try:
    if result['OK']:
      __printDictionary(result['Value'], offset=offset, shift=shift, empty=empty, depth=depth)
      return 0
    gLogger.notice("Error in %s :" % script, result['Message'])
    return 2
  except Exception as e:  # pylint: disable=broad-except
    gLogger.notice("Exception while printing results in %s - Results:" % script, repr(e))
    gLogger.notice(result)
    return 2


class ProgressBar(object):
  """This object prints a title and a progress bar on stderr."""

  def __init__(self, items, width=None, title=None, chunk=None, step=None, interactive=None, log=None):
    """

    :param items: Number of items to enumerate
    :type items: integer
    :param width: width in co.lumns of the progress bar
    :type width: int
    :param title: Title of the progress bar
    :type title: string
    :param chunk: incremental value by which the counter is incremented (in number of items)
    :type chunk: int
    :param step: frequency at which the bar is updated
    :type step: int
    :param interactive: if False, no printout
    :type interactive: boolean
    """
    if title is None:
      title = ''
    else:
      title += ' '
    if width is None:
      width = 50
    if chunk is None:
      chunk = 1
    if step is None:
      step = 1
    if interactive is None:
      interactive = True
    self._log = log if not interactive else None
    self._step = step
    self._width = width
    self._loopNumber = 0
    self._itemCounter = 0
    self._items = items
    self._chunk = chunk
    self._startTime = time.time()
    self._progress = 0
    self._showBar = bool(items > chunk) and bool(items > step) and interactive and sys.stderr.isatty()
    self._interactive = interactive
    self._title = title
    self._backspace = 0
    self._writeTitle()

  def _writeTitle(self):
    """Write the progress bar title to stderr."""
    if not self._interactive:
      return
    if self._showBar:
      sys.stderr.write("%s|%s|" % (self._title, self._width * ' '))
      self._backspace = self._width + 1
    else:
      sys.stderr.write(self._title)
    sys.stderr.flush()

  def loop(self, increment=True):
    """Called at each iteration of the loop If the iteration modulo "step" is
    0, update the bar Increment the counter of items by "chunk"."""
    if not self._interactive:
      return
    showBar = self._showBar and (self._loopNumber % self._step) == 0
    fraction = min(float(self._itemCounter) / float(self._items), 1.)
    if increment:
      self._loopNumber += 1
      self._itemCounter += self._chunk
    else:
      showBar = self._showBar
    if showBar:
      progress = int(round(self._width * fraction))
      elapsed = time.time() - self._startTime
      if elapsed > 30. and fraction:
        rest = int(elapsed * (1 - fraction) / fraction)
        estimate = ' (%d seconds left)' % rest
      else:
        estimate = ''
      blockBlue = '\033[46m'
      endblock = '\033[0m'
      sys.stderr.write("%s%s%s| %5.1f%%%s\033[K" % (self._backspace * '\b',
                                                    blockBlue + (progress - self._progress) * ' ' + endblock,
                                                    (self._width - progress) * ' ',
                                                    100. * fraction,
                                                    estimate))
      self._backspace = self._width + 8 - progress + len(estimate)
      self._progress = progress
      sys.stderr.flush()

  def endLoop(self, message=None, timing=True):
    """Closes the progress bar, printing a message and/or the timing since the
    bar was created."""
    if message is None:
      message = 'completed'
    timingMsg = ' in %.1f seconds' % (time.time() - self._startTime)
    if not self._interactive:
      # Print out message if defined
      if self._log:
        self._log("%s %s %s" % (self._title, message, timingMsg))
      return
    if self._showBar:
      sys.stderr.write("%s\033[K: %s" % ((self._progress + self._backspace + 1) * '\b',
                                         message))
    if timing:
      if not self._showBar:
        sys.stderr.write(': %s' % message)
      sys.stderr.write(timingMsg)
    sys.stderr.write('\n')
    sys.stderr.flush()

  def comment(self, message, optMsg=''):
    """Print a comment."""
    fullMsg = '\n' + message + ' %s' % optMsg if optMsg else ''
    gLogger.notice(fullMsg)
    self._writeTitle()
    self.loop(increment=False)


class DMScript(object):
  """DMScript is a class that creates default switches for DM scripts, decodes
  them and sets flags."""

  def __init__(self):
    """c'tor."""
    self.bkClient = BookkeepingClient()
    self.exceptFileTypes = []
    self.bkClientQuery = None
    self.bkClientQueryDict = {}
    self.options = {}
    self.lastFile = os.path.join(tempfile.gettempdir(), '%d.lastLFNs' % os.getppid())
    self.setLastFile = False
    self.voName = None

  def __voName(self):
    """Returns the name of the VO."""
    if self.voName is None:
      self.voName = gConfig.getValue('/DIRAC/VirtualOrganization', '')
    gLogger.verbose('VO', self.voName)
    return self.voName

  def registerDMSwitches(self):
    """Register switches related to data management, including BK."""
    self.registerBKSwitches()
    self.registerNamespaceSwitches()
    self.registerSiteSwitches()
    self.registerFileSwitches()

  def registerBKSwitches(self):
    """Register switches related to bookkeeping."""
    # BK query switches
    Script.registerSwitch("B:", "BKQuery=", "   Bookkeeping query path", self.setBKQuery)
    Script.registerSwitch("f:", "FileType=",
                          "   File type (comma separated list, to be used with --Production) [All]", self.setFileType)
    Script.registerSwitch('', "ExceptFileType=",
                          "   Exclude the (list of) file types when all are requested", self.setExceptFileTypes)
    Script.registerSwitch('', 'EventType=', '   Event type', self.setEventType)
    Script.registerSwitch("r:", "Runs=", "   Run or range of runs (r1:r2)", self.setRuns)
    Script.registerSwitch("P:", "Productions=",
                          "   Production ID to search (comma separated list)", self.setProductions)
    Script.registerSwitch('', "DQFlags=", "   DQ flag used in query", self.setDQFlags)
    Script.registerSwitch('', "StartDate=", "   Start date for the BK query", self.setStartDate)
    Script.registerSwitch('', "EndDate=", "   End date for the BK query", self.setEndDate)
    Script.registerSwitch('', "Visibility=", "   Required visibility (Yes, No, All) [Yes]", self.setVisibility)
    Script.registerSwitch('', 'ReplicaFlag=', '   Required replica flag (Yes, No, All) [Yes]', self.setReplicaFlag)
    Script.registerSwitch('', 'TCK=', '   Get files with a given TCK', self.setTCK)

  def registerNamespaceSwitches(self, action='search [ALL]'):
    """Register namespace switches."""
    Script.registerSwitch("D:", "Directory=", "   Directory to " + action, self.setDirectory)

  def registerSiteSwitches(self):
    """SE switches."""
    Script.registerSwitch("g:", "Sites=", "  Sites to consider [ALL] (comma separated list)", self.setSites)
    Script.registerSwitch("S:", "SEs=", "  SEs to consider [ALL] (comma separated list)", self.setSEs)

  def registerFileSwitches(self):
    """File switches."""
    Script.registerSwitch("", "File=", "File containing list of LFNs", self.setLFNsFromFile)
    Script.registerSwitch("l:", "LFNs=", "List of LFNs (comma separated)", self.setLFNs)
    Script.registerSwitch("", "Terminal", "LFNs are entered from stdin (--File /dev/stdin)", self.setLFNsFromTerm)
    Script.registerSwitch("", "LastLFNs", "Use last set of LFNs", self.setLFNsFromLast)

  def registerJobsSwitches(self):
    """Job switches."""
    Script.registerSwitch("", "File=", "File containing list of DIRAC jobIds", self.setJobidsFromFile)
    Script.registerSwitch("j:", "DIRACJobids=", "List of DIRAC Jobids (comma separated)", self.setJobids)
    Script.registerSwitch("", "Terminal",
                          "DIRAC Jobids are entered from stdin (--File /dev/stdin)", self.setJobidsFromTerm)

  def setProductions(self, arg):
    """Parse production numbers."""
    prods = []
    if arg.upper() == "ALL":
      self.options['Productions'] = arg
      return DIRAC.S_OK()
    try:
      for prod in arg.split(','):
        if prod.find(":") > 0:
          pr = prod.split(":")
          for i in range(int(pr[0]), int(pr[1]) + 1):
            prods.append(i)
        else:
          prods.append(prod)
      self.options['Productions'] = [int(prod) for prod in prods]
    except ValueError as e:
      gLogger.error("Invalid production switch value: %s" % arg, repr(e))
      self.options['Productions'] = ['Invalid']
      return DIRAC.S_ERROR()
    return DIRAC.S_OK()

  def setStartDate(self, arg):
    """Setter."""
    self.options['StartDate'] = arg
    return DIRAC.S_OK()

  def setEndDate(self, arg):
    """Setter."""
    self.options['EndDate'] = arg
    return DIRAC.S_OK()

  def setFileType(self, arg):
    """Setter."""
    fileTypes = arg.split(',')
    self.options['FileType'] = fileTypes
    return DIRAC.S_OK()

  def setEventType(self, arg):
    """Setter."""
    eventTypes = arg.split(',')
    self.options['EventType'] = eventTypes
    return DIRAC.S_OK()

  def setExceptFileTypes(self, arg):
    """Setter."""
    self.exceptFileTypes += arg.split(',')
    return DIRAC.S_OK()

  def setBKQuery(self, arg):
    """Setter."""
    # BKQuery could either be a BK path or a file path that contains the BK items
    self.bkClientQuery = None
    self.bkClientQueryDict = {}
    self.options['BKPath'] = arg
    return DIRAC.S_OK()

  def setRuns(self, arg):
    """Setter."""
    self.options['Runs'] = arg
    return DIRAC.S_OK()

  def setDQFlags(self, arg):
    """Setter."""
    dqFlags = arg.split(',')
    self.options['DQFlags'] = dqFlags
    return DIRAC.S_OK()

  def setTCK(self, arg):
    """Setter."""
    tcks = arg.split(',')
    self.options['TCK'] = tcks
    return DIRAC.S_OK()

  def setVisibility(self, arg):
    """Setter."""
    if arg.lower() in ('yes', 'no', 'all'):
      self.options['Visibility'] = arg
    else:
      gLogger.error('Unknown visibility flag: %s' % arg)
      return DIRAC.exit(1)
    return DIRAC.S_OK()

  def setReplicaFlag(self, arg):
    """Setter."""
    if arg.lower() in ('yes', 'no', 'all'):
      self.options['ReplicaFlag'] = arg
    else:
      gLogger.error('Unknown replica flag: %s' % arg)
      return DIRAC.exit(1)
    return DIRAC.S_OK()

  def setDirectory(self, arg):
    """Setter."""
    if os.path.exists(arg) and not os.path.isdir(arg):
      with open(arg, 'rt') as inFile:
        directories = [line.split()[0] for line in inFile.read().splitlines() if line.strip()]
    else:
      directories = arg.split(',')
    self.options.setdefault('Directory', set()).update(directories)
    return DIRAC.S_OK()

  def setSites(self, arg):
    """Setter."""
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
    try:
      siteShortNames = DMSHelpers().getShortSiteNames(withStorage=False, tier=(0, 1))
    except AttributeError:
      siteShortNames = {'CERN': 'LCG.CERN.cern', 'CNAF': 'LCG.CNAF.it', 'GRIDKA': 'LCG.GRIDKA.de',
                        'NIKHEF': 'LCG.NIKHEF.nl', 'SARA': 'LCG.SARA.nl', 'PIC': 'LCG.PIC.es',
                        'RAL': 'LCG.RAL.uk', 'IN2P3': 'LCG.IN2P3.fr', 'RRCKI': 'LCG.RRCKI.ru'}
    sites = arg.split(',')
    self.options['Sites'] = [siteShortNames.get(site.upper(), site) for site in sites]
    return DIRAC.S_OK()

  def setSEs(self, arg):
    """Setter."""
    self.options['SEs'] = arg.split(',')
    return DIRAC.S_OK()

  def setLFNs(self, arg):
    """Setter."""
    if arg:
      self.options.setdefault('LFNs', set()).update(arg.split(','))
    return DIRAC.S_OK()

  def setLFNsFromTerm(self, arg=None):
    """Setter."""
    return self.setLFNsFromFile(arg)

  def getLFNsFromList(self, lfns, directories=False):
    """Returns a list of LFNs from a list of strings LFNs start with the last
    occurence of /<vo>/ in the file name and ends with a set of delimiters If
    directories is True, only normalized directories (ending with a "/" are
    returned."""
    if isinstance(lfns, six.string_types):
      lfnList = lfns.strip().split(',')
    elif isinstance(lfns, (list, set, dict)):
      lfnList = [lfn.strip() for lfn1 in lfns for lfn in lfn1.split(',')]
    else:
      gLogger.error('getLFNsFromList: invalid type %s' % type(lfns))
      return []
    vo = self.__voName()
    if vo:
      vo = '/%s' % vo
      lfnList = [x.split('LFN:')[-1].strip() for x in lfnList]
      for sep in ('"', ',', "'", ':', '(', ')', ';', '|'):
        lfnList = [x.replace(sep, ' ') for x in lfnList]
      lfnList = [vo + lfn.split(vo)[-1].split()[0]
                 if vo in lfn
                 else lfn if lfn == vo else ''
                 for lfn in lfnList]
      lfnList = [lfn.split('?')[0] for lfn in lfnList]
      lfnList = [lfn for lfn in lfnList if lfn.endswith('/')] if directories else \
                [lfn for lfn in lfnList if not lfn.endswith('/')]
    return sorted(lfn for lfn in set(lfnList) if lfn)

  @staticmethod
  def getJobIDsFromList(jobids):
    """it returns a list of jobids using a string."""
    jobidsList = []
    if isinstance(jobids, six.string_types):
      jobidsList = jobids.split(',')
    elif isinstance(jobids, list):
      jobidsList = [jobid for jobid1 in jobids for jobid in jobid1.split(',')]
    jobidsList = [jobid for jobid in jobidsList if jobid]
    return jobidsList

  def setLFNsFromLast(self, _val):
    """Setter when --Last is used."""
    if os.path.exists(self.lastFile):
      return self.setLFNsFromFile(self.lastFile)
    gLogger.fatal('Last file %s does not exist' % self.lastFile)
    DIRAC.exit(2)

  def setLFNsFromFile(self, arg):
    """Reads the content of a file or from stdin (in which case a temporary
    file will be created) LFNs are not parsed at this stage."""
    if isinstance(arg, six.string_types) and arg.lower() == 'last':
      arg = self.lastFile
    # Make a list of files
    if isinstance(arg, six.string_types):
      files = arg.split(',')
    elif isinstance(arg, list):
      files = arg
    elif not arg:
      files = [arg]
    nfiles = 0
    for fName in files:
      try:
        with open(fName if fName else '/dev/stdin', 'rt') as inFile:
          lfns = inFile.read().splitlines()
        nfiles += len(lfns)
      except (EOFError, IOError):
        lfns = fName.split(',')
      self.options.setdefault('LFNs', set()).update(lfns)
    if nfiles:
      self.setLastFile = arg if arg else 'term'
    return DIRAC.S_OK()

  def getOptions(self):
    """Returns all options."""
    return self.options

  def getOption(self, switch, default=None):
    """Get a specific items set by the setters."""
    if switch == 'SEs':
      # SEs have to be resolved recursively using StorageElementGroups
      return resolveSEGroup(self.options.get(switch, default))
    value = self.options.get(switch, default)
    if switch in ('LFNs', 'Directory'):
      # Special case for getting LFNs or directories: parse the option
      if value == default and switch == 'Directory':
        value = self.options.get('LFNs', default)
      if not value:
        if not sys.stdin.isatty():
          # If the input file is a pipe, no need to specify it
          self.setLFNsFromTerm()
          value = self.options.get('LFNs', default)
      if value:
        # Parse the LFNs out of the "LFNs" option list
        value = self.getLFNsFromList(value, directories=switch == 'Directory')
      if value and self.setLastFile and switch == 'LFNs':
        # Storethe list of LFNs in a temporary file
        gLogger.always("Got %d LFNs" % len(value))
        if self.setLastFile != self.lastFile:
          self. setLastFile = False
          with open(self.lastFile, 'wt') as tmpFile:
            tmpFile.write('\n'.join(sorted(value)))
    if isinstance(value, set):
      # Return a sorted list from a set
      value = sorted(value)
    return value

  def getBKQuery(self, visible=None):
    """Returns a BKQuery object from the requested BK information."""
    mandatoryKeys = {('ConfigName', 'ConfigVersion'),
                     'Production',
                     ('FileType', 'RunNumber'),
                     ('FileType', 'StartRun')}
    if self.bkClientQuery:
      return self.bkClientQuery
    if self.bkClientQueryDict:
      self.bkClientQuery = BKQuery(self.bkClientQueryDict)
    else:
      visible = self.options.get('Visibility', 'Yes') if visible is None else visible
      bkPath = self.options.get('BKPath')
      prods = self.options.get('Productions')
      runs = self.options.get('Runs')
      fileTypes = self.options.get('FileType')
      eventTypes = self.options.get('EventType')
      self.bkClientQuery = BKQuery(bkQuery=bkPath, prods=prods, runs=runs,
                                   fileTypes=fileTypes, eventTypes=eventTypes, visible=visible)
    bkQueryDict = self.bkClientQuery.getQueryDict()
    found = False
    for key in mandatoryKeys:
      if isinstance(key, six.string_types) and key in bkQueryDict:
        found = True
        break
      elif isinstance(key, (list, tuple)) and not set(key) - set(bkQueryDict):
        found = True
        break
    if not found:
      self.bkClientQuery = None
      return None
    # Add extra requirements
    self.bkClientQuery.setExceptFileTypes(self.exceptFileTypes)
    if 'DQFlags' in self.options:
      self.bkClientQuery.setDQFlag(self.options['DQFlags'])
    if 'StartDate' in self.options:
      self.bkClientQuery.setOption('StartDate', self.options['StartDate'])
    if 'EndDate' in self.options:
      self.bkClientQuery.setOption('EndDate', self.options['EndDate'])
    if 'ReplicaFlag' in self.options:
      self.bkClientQuery.setOption('ReplicaFlag', self.options['ReplicaFlag'])
    if 'TCK' in self.options:
      self.bkClientQuery.setOption('TCK', self.options['TCK'])
    return self.bkClientQuery

  def getRequestID(self, prod=None):
    """Get the request ID for a single production."""
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    if not prod:
      prod = self.options.get('Productions', [])
    requestID = None
    if isinstance(prod, six.string_types):
      prods = [prod]
    else:
      prods = prod
    if len(prods) == 1 and str(prods[0]).upper() != 'ALL':
      res = TransformationClient().getTransformation(prods[0])
      if res['OK']:
        requestID = int(res['Value']['TransformationFamily'])
    return requestID

  def setJobidsFromFile(self, arg):
    """It fill a list with DIRAC jobids read from a file
    NOTE: The file format is equivalent to the file format when the content is
    a list of LFNs."""
    try:
      with open(arg if arg else '/dev/stdin', 'rt') as inFile:
        jobids = self.getJobIDsFromList(inFile.read().splitlines())
    except IOError as error:
      gLogger.exception('Reading jobids from a file is failed with exception:', repr(error))
      jobids = self.getJobIDsFromList(arg)
    self.options.setdefault('JobIDs', set()).update(jobids)
    return DIRAC.S_OK()

  def setJobids(self, arg):
    """It fill a list with DIRAC Jobids."""
    if arg:
      self.options.setdefault('JobIDs', set()).update(self.getJobIDsFromList(arg))
    return DIRAC.S_OK()

  def setJobidsFromTerm(self):
    """It is used to fill a list with jobids."""
    return self.setJobidsFromFile(None)
